import torch
import tensorrt as trt
from .flattener import Flattener
from .misc_utils import (
    torch_dtype_from_trt,
    torch_device_from_trt
)
from .version_utils import (
    trt_version
)


class TRTModule(torch.nn.Module):
    def __init__(self, engine=None, input_names=None, output_names=None, input_flattener=None, output_flattener=None):
        super(TRTModule, self).__init__()
        self._register_state_dict_hook(TRTModule._on_state_dict)

        if isinstance(engine, str):
            # assume filepath
            with open(engine, 'rb') as f:
                engine = f.read()
            with trt.Logger() as logger, trt.Runtime(logger) as runtime:
                engine = runtime.deserialize_cuda_engine(engine)
        elif isinstance(engine, trt.IHostMemory):
            with trt.Logger() as logger, trt.Runtime(logger) as runtime:
                engine = runtime.deserialize_cuda_engine(engine)
                
        self.engine = engine
        if self.engine is not None:
            self.context = self.engine.create_execution_context()
            self._update_name_binindgs_maps()
        self.input_names = input_names
        self.output_names = output_names
        self.input_flattener = input_flattener
        self.output_flattener = output_flattener
    
    def _update_name_binindgs_maps(self):
        if trt_version() >= "10.0":
            self._update_name_binding_maps_trt_10()
        else:
            self._update_name_binding_maps_pre_trt_10()

    def _update_name_binding_maps_trt_10(self):
        self._name_to_binding = {}
        self._binding_to_name = {}
        for i in range(self.engine.num_io_tensors):
            name_i = self.engine.get_tensor_name(i)
            self._name_to_binding[name_i] = i
            self._binding_to_name[i] = name_i

    def _update_name_binding_maps_pre_trt_10(self):
        self._name_to_binding = {}
        self._binding_to_name = {}
        for i in range(self.engine.num_bindings):
            name_i = self.engine.get_binding_name(i)
            self._name_to_binding[name_i] = i
            self._binding_to_name[i] = name_i

    def _on_state_dict(self, state_dict, prefix, local_metadata):
        state_dict[prefix + "engine"] = bytearray(self.engine.serialize())
        state_dict[prefix + "input_names"] = self.input_names
        state_dict[prefix + "output_names"] = self.output_names
        state_dict[prefix + "input_flattener"] = self.input_flattener.dict()
        state_dict[prefix + "output_flattener"] = self.output_flattener.dict()

    def _load_from_state_dict(
        self,
        state_dict,
        prefix,
        local_metadata,
        strict,
        missing_keys,
        unexpected_keys,
        error_msgs,
    ):
        engine_bytes = state_dict[prefix + "engine"]

        with trt.Logger() as logger, trt.Runtime(logger) as runtime:
            self.engine = runtime.deserialize_cuda_engine(engine_bytes)
            self.context = self.engine.create_execution_context()

        self.input_names = state_dict[prefix + "input_names"]
        self.output_names = state_dict[prefix + "output_names"]

        if 'input_flattener' in state_dict:
            self.input_flattener = Flattener.from_dict(state_dict['input_flattener'])
        else:
            self.input_flattener = None

        if 'output_flattener' in state_dict:
            self.output_flattener = Flattener.from_dict(state_dict['output_flattener'])
        else:
            self.output_flattener = None

        self._update_name_binindgs_maps()

    def _forward_pre_10(self, *inputs):
        bindings = [None] * (len(self.input_names) + len(self.output_names))
        
        if self.input_flattener is not None:
            inputs = self.input_flattener.flatten(inputs)

        for i, input_name in enumerate(self.input_names):
            idx = self.engine.get_binding_index(input_name)
            shape = tuple(inputs[i].shape)
            bindings[idx] = inputs[i].contiguous().data_ptr()
            self.context.set_binding_shape(idx, shape)

        # create output tensors
        outputs = [None] * len(self.output_names)
        for i, output_name in enumerate(self.output_names):
            idx = self.engine.get_binding_index(output_name)
            dtype = torch_dtype_from_trt(self.engine.get_binding_dtype(idx))
            shape = tuple(self.context.get_binding_shape(idx))
            device = torch_device_from_trt(self.engine.get_location(idx))
            output = torch.empty(size=shape, dtype=dtype, device=device)
            outputs[i] = output
            bindings[idx] = output.data_ptr()

        self.context.execute_async_v2(
            bindings, torch.cuda.current_stream().cuda_stream
        )

        if self.output_flattener is not None:
            outputs = self.output_flattener.unflatten(outputs)
        else:
            outputs = tuple(outputs)
            if len(outputs) == 1:
                outputs = outputs[0]

        return outputs

    def _forward_post_10(self, *inputs):
        if self.input_flattener is not None:
            inputs = self.input_flattener.flatten(inputs)

        # set shapes
        for i, input_name in enumerate(self.input_names):
            shape = tuple(inputs[i].shape)
            data_ptr = inputs[i].contiguous().data_ptr()
            self.context.set_tensor_address(input_name, data_ptr)
            self.context.set_input_shape(input_name, shape)

        # execute
        outputs = [None] * len(self.output_names)
        for i, output_name in enumerate(self.output_names):
            dtype = torch_dtype_from_trt(self.engine.get_tensor_dtype(output_name))
            shape = tuple(self.context.get_tensor_shape(output_name))
            device = torch_device_from_trt(self.engine.get_tensor_location(output_name))
            output = torch.empty(size=shape, dtype=dtype, device=device)
            outputs[i] = output
            self.context.set_tensor_address(output_name, output.data_ptr())

        self.context.execute_async_v3(torch.cuda.current_stream().cuda_stream)

        if self.output_flattener is not None:
            outputs = self.output_flattener.unflatten(outputs)
        else:
            outputs = tuple(outputs)
            if len(outputs) == 1:
                outputs = outputs[0]

        return outputs

    def forward(self, *inputs):
        if trt_version() < "10.0":
            return self._forward_pre_10(*inputs)
        else:
            return self._forward_post_10(*inputs)

    def enable_profiling(self):
        if not self.context.profiler:
            self.context.profiler = trt.Profiler()



import os
import torch
import glob
from uuid import uuid1
from jetson.torch2trt.flattener import Flattener


__all__ = [
    'DatasetRecorder',
    'Dataset',
    'ListDataset',
    'TensorBatchDataset'
]


class DatasetRecorder(object):

    def __init__(self, dataset, module):
        self.dataset = dataset
        self.module = module
        self.handle = None

    def __enter__(self, *args, **kwargs):

        if self.handle is not None:
            raise RuntimeError('DatasetRecorder is already active.')

        self.handle = self.module.register_forward_pre_hook(self._callback)

        return self

    def __exit__(self, *args, **kwargs):
        if self.handle is not None:
            self.handle.remove()
            self.handle = None

    def _callback(self, module, input):
        self.dataset.insert(input)


class Dataset(object):

    def __len__(self):
        raise NotImplementedError

    def __getitem__(self, index):
        raise NotImplementedError

    def insert(self, item):
        raise NotImplementedError

    def record(self, module):
        return DatasetRecorder(self, module)

    def num_inputs(self):
        return len(self.getitem_flat(0))

    @property
    def flattener(self):
        if not hasattr(self, '_flattener') or self._flattener is None:
            assert(len(self) > 0, 'Cannot create default flattener without input data.')
            value = self[0]
            self._flattener = Flattener.from_value(value)
        return self._flattener

    def getitem_flat(self, index):
        return self.flattener.flatten(self[index])
    
    def shapes_for_index(self, index, flat=False):
        shapes = [None for i in range(self.num_inputs())]
        tensors = self.getitem_flat(index)
        for j in range(len(tensors)):
            shapes[j] = torch.Size(tuple(tensors[j].shape))

        if flat:
            return shapes
        else:
            return self.flattener.unflatten(shapes)

    def shapes(self, flat=False):
        shapes = [[] for i in range(self.num_inputs())]
        for i in range(len(self)):
            tensors = self.getitem_flat(i)
            for j in range(len(tensors)):
                shapes[j].append(torch.Size(tuple(tensors[j].shape)))

        if flat:
            return shapes
        else:
            return self.flattener.unflatten(shapes)

    def _shape_stats(self, stat_fn, flat=False):
        shapes = []
        for s in self.shapes(flat=True):
            shape_tensor = []
            for si in s:
                shape_tensor.append(tuple(si))
            shape_tensor = torch.LongTensor(shape_tensor)
            shapes.append(shape_tensor)
        
        stat_shapes = []
        for shape in shapes:
            stat_shape = torch.Size(stat_fn(shape))
            stat_shapes.append(stat_shape)
        if flat:
            return stat_shapes
        else:
            return self.flattener.unflatten(stat_shapes)

    def min_shapes(self, flat=False):
        return self._shape_stats(lambda x: torch.min(x, dim=0)[0], flat)

    def max_shapes(self, flat=False):
        return self._shape_stats(lambda x: torch.max(x, dim=0)[0], flat)

    def item_numel(self, index):
        tensors = self.getitem_flat(index)
        return sum([t.numel() for t in tensors])

    def median_numel_shapes(self, flat=False):
        numels = torch.LongTensor([self.item_numel(i) for i in range(len(self))])
        median_index = int(torch.argsort(numels)[len(numels) // 2])
        return self.shapes_for_index(median_index, flat=flat)

    def infer_dynamic_axes(self, flat=False):
        min_shapes = self.min_shapes(flat=True)
        max_shapes = self.max_shapes(flat=True)
        dynamic_axes = [[] for i in range(self.num_inputs())]
        for i, (mins, maxs) in enumerate(zip(min_shapes, max_shapes)):
            for j, (mins_i, maxs_i) in enumerate(zip(mins, maxs)):
                if mins_i != maxs_i:
                    dynamic_axes[i].append(j)
        if flat:
            return dynamic_axes
        else:
            return self.flattener.unflatten(dynamic_axes)


class ListDataset(Dataset):

    def __init__(self, items=None):
        if items is None:
            items = []
        self.items = [t for t in items]

    def __len__(self):
        return len(self.items)

    def __getitem__(self, index):
        return self.items[index]

    def insert(self, item):
        self.items.append(item)


class TensorBatchDataset(Dataset):

    def __init__(self, tensors=None):
        if tensors is not None:
            self._flattener = Flattener.from_value(tensors)
            self.tensors = self._flattener.flatten(tensors)
        else:
            self._flattener = None
            self.tensors = None

    def __len__(self):
        if self.tensors is None:
            return 0
        else:
            return len(self.tensors[0])

    def __getitem__(self, idx):
        if self.tensors is None:
            raise IndexError('Dataset is empty.')
        return self.flattener.unflatten([t[idx:idx+1] for t in self.tensors])

    def insert(self, tensors):
        if self._flattener is None:
            self._flattener = Flattener.from_value(tensors)

        tensors = self.flattener.flatten(tensors)

        if self.tensors is None:
            self.tensors = tensors
        else:
            if len(self.tensors) != len(tensors):
                raise ValueError('Number of inserted tensors does not match the number of tensors in the current dataset.')
            
            self.tensors = tuple([
                torch.cat((self.tensors[index], tensors[index]), dim=0) 
                for index in range(len(tensors))
            ])


class FolderDataset(Dataset):

    def __init__(self, folder):
        super().__init__()
        if not os.path.exists(folder):
            os.makedirs(folder)
        self.folder = folder
    
    def file_paths(self):
        return sorted(glob.glob(os.path.join(self.folder, '*.pth')))

    def __len__(self):
        return len(self.file_paths())

    def __getitem__(self, index):
        return torch.load(self.file_paths()[index])

    def insert(self, tensors):
        i = 0
        file_paths = [os.path.basename(path) for path in self.file_paths()]
        while ('input_%d.pth' % i) in file_paths:
            i += 1
        torch.save(tensors, os.path.join(self.folder, 'input_%d.pth' % i))


"""plugins.py

I referenced the code from https://github.com/dongfangduoshou123/YoloV3-TensorRT/blob/master/seralizeEngineFromPythonAPI.py
"""


import ctypes

import numpy as np
import tensorrt as trt
import os

from yolo_to_onnx import (is_pan_arch, DarkNetParser, get_category_num,
                          get_h_and_w, get_output_convs, get_anchors)

lib_path = os.path.abspath(__file__ + "/../../models/yolov4-tiny/libyolo_layer.so")

try:
    ctypes.cdll.LoadLibrary(lib_path)
except OSError as e:
    raise SystemExit(f'ERROR: failed to load {lib_path}.  '
                     'Did you forget to do a "make" in the "../plugins/" '
                     'subdirectory?') from e


def get_scales(cfg_file_path):
    """Get scale_x_y's of all yolo layers from the cfg file."""
    with open(cfg_file_path, 'r') as f:
        cfg_lines = f.readlines()
    yolo_lines = [l.strip() for l in cfg_lines if l.startswith('[yolo]')]
    scale_lines = [l.strip() for l in cfg_lines if l.startswith('scale_x_y')]
    if len(scale_lines) == 0:
        return [1.0] * len(yolo_lines)
    else:
        assert len(scale_lines) == len(yolo_lines)
        return [float(l.split('=')[-1]) for l in scale_lines]


def get_new_coords(cfg_file_path):
    """Get new_coords flag of yolo layers from the cfg file."""
    with open(cfg_file_path, 'r') as f:
        cfg_lines = f.readlines()
    yolo_lines = [l.strip() for l in cfg_lines if l.startswith('[yolo]')]
    newc_lines = [l.strip() for l in cfg_lines if l.startswith('new_coords')]
    if len(newc_lines) == 0:
        return 0
    else:
        assert len(newc_lines) == len(yolo_lines)
        return int(newc_lines[-1].split('=')[-1])


def get_plugin_creator(plugin_name, logger):
    """Get the TensorRT plugin creator."""
    trt.init_libnvinfer_plugins(logger, '')
    plugin_creator_list = trt.get_plugin_registry().plugin_creator_list
    for c in plugin_creator_list:
        if c.name == plugin_name:
            return c
    return None


def add_yolo_plugins(network, model_name, logger):
    """Add yolo plugins into a TensorRT network."""
    cfg_file_path = model_name + '.cfg'
    parser = DarkNetParser()
    layer_configs = parser.parse_cfg_file(cfg_file_path)
    num_classes = get_category_num(cfg_file_path)
    output_tensor_names = get_output_convs(layer_configs)
    h, w = get_h_and_w(layer_configs)
    if len(output_tensor_names) == 2:
        yolo_whs = [
            [w // 32, h // 32], [w // 16, h // 16]]
    elif len(output_tensor_names) == 3:
        yolo_whs = [
            [w // 32, h // 32], [w // 16, h // 16],
            [w //  8, h //  8]]
    elif len(output_tensor_names) == 4:
        yolo_whs = [
            [w // 64, h // 64], [w // 32, h // 32],
            [w // 16, h // 16], [w //  8, h //  8]]
    else:
        raise TypeError('bad number of outputs: %d' % len(output_tensor_names))
    if is_pan_arch(cfg_file_path):
        yolo_whs.reverse()
    anchors = get_anchors(cfg_file_path)
    if len(anchors) != len(yolo_whs):
        raise ValueError('bad number of yolo layers: %d vs. %d' %
                         (len(anchors), len(yolo_whs)))
    if network.num_outputs != len(anchors):
        raise ValueError('bad number of network outputs: %d vs. %d' %
                         (network.num_outputs, len(anchors)))
    scales = get_scales(cfg_file_path)
    if any([s < 1.0 for s in scales]):
        raise ValueError('bad scale_x_y: %s' % str(scales))
    if len(scales) != len(anchors):
        raise ValueError('bad number of scales: %d vs. %d' %
                         (len(scales), len(anchors)))
    new_coords = get_new_coords(cfg_file_path)

    plugin_creator = get_plugin_creator('YoloLayer_TRT', logger)
    if not plugin_creator:
        raise RuntimeError('cannot get YoloLayer_TRT plugin creator')
    old_tensors = [network.get_output(i) for i in range(network.num_outputs)]
    new_tensors = [None] * network.num_outputs
    for i, old_tensor in enumerate(old_tensors):
        input_multiplier = w // yolo_whs[i][0]
        new_tensors[i] = network.add_plugin_v2(
            [old_tensor],
            plugin_creator.create_plugin('YoloLayer_TRT', trt.PluginFieldCollection([
                trt.PluginField("yoloWidth", np.array(yolo_whs[i][0], dtype=np.int32), trt.PluginFieldType.INT32),
                trt.PluginField("yoloHeight", np.array(yolo_whs[i][1], dtype=np.int32), trt.PluginFieldType.INT32),
                trt.PluginField("inputMultiplier", np.array(input_multiplier, dtype=np.int32), trt.PluginFieldType.INT32),
                trt.PluginField("newCoords", np.array(new_coords, dtype=np.int32), trt.PluginFieldType.INT32),
                trt.PluginField("numClasses", np.array(num_classes, dtype=np.int32), trt.PluginFieldType.INT32),
                trt.PluginField("numAnchors", np.array(len(anchors[i]) // 2, dtype=np.int32), trt.PluginFieldType.INT32),
                trt.PluginField("anchors", np.ascontiguousarray(anchors[i], dtype=np.float32), trt.PluginFieldType.FLOAT32),
                trt.PluginField("scaleXY", np.array(scales[i], dtype=np.float32), trt.PluginFieldType.FLOAT32),
            ]))
        ).get_output(0)

    for new_tensor in new_tensors:
        network.mark_output(new_tensor)
    for old_tensor in old_tensors:
        network.unmark_output(old_tensor)

    return network


def add_concat(network, model_name, logger):
    """Add a final concatenation output into a TensorRT network."""
    if network.num_outputs < 2 or network.num_outputs > 4:
        raise TypeError('bad number of yolo layers: %d' % network.num_outputs)
    yolo_tensors = [network.get_output(i) for i in range(network.num_outputs)]
    concat_tensor = network.add_concatenation(yolo_tensors).get_output(0)
    for yolo_tensor in yolo_tensors:
        network.unmark_output(yolo_tensor)
    concat_tensor.name = 'detections'
    network.mark_output(concat_tensor)
    return network



import os
import sys
import argparse
import tensorrt as trt

path = os.path.abspath(__file__+"/..")        
sys.path.append(path)

from yolo_to_onnx import DarkNetParser, get_h_and_w
from plugins import add_yolo_plugins, add_concat


MAX_BATCH_SIZE = 1


def get_c(layer_configs):
    """Find input channels of the yolo model from layer configs."""
    net_config = layer_configs['000_net']
    return net_config.get('channels', 3)


def load_onnx(model_name):
    """Read the ONNX file."""
    onnx_path = '%s.onnx' % model_name
    if not os.path.isfile(onnx_path):
        print('ERROR: file (%s) not found!  You might want to run yolo_to_onnx.py first to generate it.' % onnx_path)
        return None
    else:
        with open(onnx_path, 'rb') as f:
            return f.read()


def set_net_batch(network, batch_size):
    """Set network input batch size.

    The ONNX file might have been generated with a different batch size,
    say, 64.
    """
    if trt.__version__[0] >= '7':
        shape = list(network.get_input(0).shape)
        shape[0] = batch_size
        network.get_input(0).shape = shape
    return network


def build_engine(model_name, do_int8, dla_core, verbose=False):
    """Build a TensorRT engine from ONNX using the older API."""
    cfg_file_path = model_name + '.cfg'
    parser = DarkNetParser()
    layer_configs = parser.parse_cfg_file(cfg_file_path)
    net_c = get_c(layer_configs)
    net_h, net_w = get_h_and_w(layer_configs)

    print('Loading the ONNX file...')
    onnx_data = load_onnx(model_name)
    if onnx_data is None:
        return None

    TRT_LOGGER = trt.Logger(trt.Logger.VERBOSE) if verbose else trt.Logger()
    EXPLICIT_BATCH = [] if trt.__version__[0] < '7' else \
        [1 << (int)(trt.NetworkDefinitionCreationFlag.EXPLICIT_BATCH)]
    with trt.Builder(TRT_LOGGER) as builder, builder.create_network(*EXPLICIT_BATCH) as network, trt.OnnxParser(network, TRT_LOGGER) as parser:
        if do_int8 and not builder.platform_has_fast_int8:
            raise RuntimeError('INT8 not supported on this platform')
        if not parser.parse(onnx_data):
            print('ERROR: Failed to parse the ONNX file.')
            for error in range(parser.num_errors):
                print(parser.get_error(error))
            return None
        network = set_net_batch(network, MAX_BATCH_SIZE)

        print('Adding yolo_layer plugins.')
        network = add_yolo_plugins(network, model_name, TRT_LOGGER)

        print('Adding a concatenated output as "detections".')
        network = add_concat(network, model_name, TRT_LOGGER)

        print('Naming the input tensort as "input".')
        network.get_input(0).name = 'input'

        print('Building the TensorRT engine.  This would take a while...')
        print('(Use "--verbose" or "-v" to enable verbose logging.)')
        if trt.__version__[0] < '7':  # older API: build_cuda_engine()
            if dla_core >= 0:
                raise RuntimeError('DLA core not supported by old API')
            builder.max_batch_size = MAX_BATCH_SIZE
            builder.max_workspace_size = 1 << 30
            builder.fp16_mode = True  # alternative: builder.platform_has_fast_fp16
            if do_int8:
                from calibrator import YOLOEntropyCalibrator
                builder.int8_mode = True
                builder.int8_calibrator = YOLOEntropyCalibrator(
                    'calib_images', (net_h, net_w), 'calib_%s.bin' % model_name)
            engine = builder.build_cuda_engine(network)
        else:  # new API: build_engine() with builder config
            builder.max_batch_size = MAX_BATCH_SIZE
            config = builder.create_builder_config()
            config.max_workspace_size = 1 << 30
            config.set_flag(trt.BuilderFlag.GPU_FALLBACK)
            config.set_flag(trt.BuilderFlag.FP16)
            profile = builder.create_optimization_profile()
            profile.set_shape(
                'input',                                # input tensor name
                (MAX_BATCH_SIZE, net_c, net_h, net_w),  # min shape
                (MAX_BATCH_SIZE, net_c, net_h, net_w),  # opt shape
                (MAX_BATCH_SIZE, net_c, net_h, net_w))  # max shape
            config.add_optimization_profile(profile)
            if do_int8:
                from calibrator import YOLOEntropyCalibrator
                config.set_flag(trt.BuilderFlag.INT8)
                config.int8_calibrator = YOLOEntropyCalibrator(
                    'calib_images', (net_h, net_w),
                    'calib_%s.bin' % model_name)
                config.set_calibration_profile(profile)
            if dla_core >= 0:
                config.default_device_type = trt.DeviceType.DLA
                config.DLA_core = dla_core
                config.set_flag(trt.BuilderFlag.STRICT_TYPES)
                print('Using DLA core %d.' % dla_core)
            engine = builder.build_engine(network, config)

        if engine is not None:
            print('Completed creating engine.')
        return engine


def convert(model, int8=False, dla_core=-1, verbose=False):
    engine = build_engine(
        model, int8, dla_core, verbose)
    if engine is None:
        raise SystemExit('ERROR: failed to build the TensorRT engine!')

    engine_path = '%s.trt' % model
    with open(engine_path, 'wb') as f:
        f.write(engine.serialize())
    print('Serialized the TensorRT engine to file: %s' % engine_path)



from .torch2trt import *
from .converters import *
import tensorrt as trt

def load_plugins():
    import jetson.torch2trt.torch_plugins
    registry = trt.get_plugin_registry()
    torch2trt_creators = [c for c in registry.plugin_creator_list if c.plugin_namespace == 'torch2trt']
    for c in torch2trt_creators:
        registry.register_creator(c, 'torch2trt')

try:
    load_plugins()
except:
    pass



import packaging.version
import tensorrt as trt
import torch


def trt_version():
    return Version(trt.__version__)


def torch_version():
    return Version(torch.__version__)


class Version(packaging.version.Version):

    def __ge__(self, other):
        if isinstance(other, str):
            other = Version(other)
        return super().__ge__(other)

    def __le__(self, other):
        if isinstance(other, str):
            other = Version(other)
        return super().__le__(other)

    def __eq__(self, other):
        if isinstance(other, str):
            other = Version(other)
        return super().__eq__(other)

    def __gt__(self, other):
        if isinstance(other, str):
            other = Version(other)
        return super().__gt__(other)
    
    def __lt__(self, other):
        if isinstance(other, str):
            other = Version(other)
        return super().__lt__(other)
    


#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''
# ported from https://gist.github.com/f0k/63a664160d016a491b2cbea15913d549
'''

import ctypes

CUDA_SUCCESS = 0

def get_gpu_archs():
    libnames = ('libcuda.so', 'libcuda.dylib', 'cuda.dll')
    for libname in libnames:
        try:
            cuda = ctypes.CDLL(libname)
        except OSError:
            continue
        else:
            break
    else:
        return

    gpu_archs = set()

    n_gpus = ctypes.c_int()
    cc_major = ctypes.c_int()
    cc_minor = ctypes.c_int()

    result = ctypes.c_int()
    device = ctypes.c_int()
    error_str = ctypes.c_char_p()

    result = cuda.cuInit(0)
    if result != CUDA_SUCCESS:
        cuda.cuGetErrorString(result, ctypes.byref(error_str))
        # print('cuInit failed with error code %d: %s' % (result, error_str.value.decode()))
        return []

    result = cuda.cuDeviceGetCount(ctypes.byref(n_gpus))
    if result != CUDA_SUCCESS:
        cuda.cuGetErrorString(result, ctypes.byref(error_str))
        # print('cuDeviceGetCount failed with error code %d: %s' % (result, error_str.value.decode()))
        return []

    for i in range(n_gpus.value):
        if cuda.cuDeviceComputeCapability(ctypes.byref(cc_major), ctypes.byref(cc_minor), device) == CUDA_SUCCESS:
            gpu_archs.add(str(cc_major.value) + str(cc_minor.value))

    return list(gpu_archs)

if __name__ == '__main__':
    print(' '.join(get_gpu_archs()))



import torch
import torchvision
from .common import *


class MnasnetBackbone(torch.nn.Module):
    
    def __init__(self, backbone):
        super(MnasnetBackbone, self).__init__()
        self.backbone = backbone
    
    def forward(self, x):
        x = self.backbone.layers(x)
        return x
    
    
def _mnasnet_pose(cmap_channels, paf_channels, upsample_channels, backbone, feature_channels, num_upsample, num_flat):
    model = torch.nn.Sequential(
        MnasnetBackbone(backbone),
        CmapPafHead(feature_channels, cmap_channels, paf_channels, upsample_channels, num_upsample=num_upsample, num_flat=num_flat)
    )
    return model
    
    
def _mnasnet_pose_att(cmap_channels, paf_channels, upsample_channels, backbone, feature_channels, num_upsample, num_flat):
    model = torch.nn.Sequential(
        MnasnetBackbone(backbone),
        CmapPafHeadAttention(feature_channels, cmap_channels, paf_channels, upsample_channels, num_upsample=num_upsample, num_flat=num_flat)
    )
    return model


def mnasnet0_5_baseline(cmap_channels, paf_channels, upsample_channels=256, pretrained=True, num_upsample=3, num_flat=0):
    backbone = torchvision.models.mnasnet0_5(pretrained=pretrained)
    return _mnasnet_pose(cmap_channels, paf_channels, upsample_channels, backbone, 1280, num_upsample, num_flat)


def mnasnet0_75_baseline(cmap_channels, paf_channels, upsample_channels=256, pretrained=True, num_upsample=3, num_flat=0):
    backbone = torchvision.models.mnasnet0_75(pretrained=pretrained)
    return _mnasnet_pose(cmap_channels, paf_channels, upsample_channels, backbone, 1280, num_upsample, num_flat)


def mnasnet1_0_baseline(cmap_channels, paf_channels, upsample_channels=256, pretrained=True, num_upsample=3, num_flat=0):
    backbone = torchvision.models.mnasnet1_0(pretrained=pretrained)
    return _mnasnet_pose(cmap_channels, paf_channels, upsample_channels, backbone, 1280, num_upsample, num_flat)


def mnasnet1_3_baseline(cmap_channels, paf_channels, upsample_channels=256, pretrained=True, num_upsample=3, num_flat=0):
    backbone = torchvision.models.mnasnet1_3(pretrained=pretrained)
    return _mnasnet_pose(cmap_channels, paf_channels, upsample_channels, backbone, 1280, num_upsample, num_flat)



def mnasnet0_5_baseline_att(cmap_channels, paf_channels, upsample_channels=256, pretrained=True, num_upsample=3, num_flat=0):
    backbone = torchvision.models.mnasnet0_5(pretrained=pretrained)
    return _mnasnet_pose_att(cmap_channels, paf_channels, upsample_channels, backbone, 1280, num_upsample, num_flat)


def mnasnet0_75_baseline_att(cmap_channels, paf_channels, upsample_channels=256, pretrained=True, num_upsample=3, num_flat=0):
    backbone = torchvision.models.mnasnet0_75(pretrained=pretrained)
    return _mnasnet_pose_att(cmap_channels, paf_channels, upsample_channels, backbone, 1280, num_upsample, num_flat)


def mnasnet1_0_baseline_att(cmap_channels, paf_channels, upsample_channels=256, pretrained=True, num_upsample=3, num_flat=0):
    backbone = torchvision.models.mnasnet1_0(pretrained=pretrained)
    return _mnasnet_pose_att(cmap_channels, paf_channels, upsample_channels, backbone, 1280, num_upsample, num_flat)


def mnasnet1_3_baseline_att(cmap_channels, paf_channels, upsample_channels=256, pretrained=True, num_upsample=3, num_flat=0):
    backbone = torchvision.models.mnasnet1_3(pretrained=pretrained)
    return _mnasnet_pose_att(cmap_channels, paf_channels, upsample_channels, backbone, 1280, num_upsample, num_flat)


from jetson.torch2trt.torch2trt import *


def is_private(method):
    method = method.split('.')[-1]  # remove prefix
    return method[0] == '_' and method[1] != '_'

def is_function_type(method):
    fntype =  eval(method + '.__class__.__name__')
    return fntype == 'function' or fntype == 'builtin_function_or_method' or fntype == 'method_descriptor'

def get_methods(namespace):
    methods = []
    for method in dir(eval(namespace)):
        full_method = namespace + '.' + method
        if not is_private(full_method) and is_function_type(full_method):
            methods.append(full_method)
    return methods


TORCH_METHODS = []
TORCH_METHODS += get_methods('torch')
TORCH_METHODS += get_methods('torch.Tensor')
TORCH_METHODS += get_methods('torch.nn.functional')


for method in TORCH_METHODS:
    
    @tensorrt_converter(method, is_real=False)
    def warn_method(ctx):
        print('Warning: Encountered known unsupported method %s' % ctx.method_str)
        

@tensorrt_converter('torch.Tensor.dim', is_real=False)
@tensorrt_converter('torch.Tensor.size', is_real=False)
def dont_warn(ctx):
    pass



import __main__


def isnotebook():
    try:
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return True
        else:
            return False
    except NameError:
        return False

__main__.isnotebook = isnotebook()


import sys
import os, subprocess


version_info = sys.version_info

def __bootstrap__():
    global __bootstrap__, __loader__, __file__
    import sys, pkg_resources, importlib.util
    path = os.path.dirname(__file__)
    lib = os.path.join(path, 'plugins.cpython-aarch64-linux-gnu.so')
    if not os.path.exists(lib):
        ori_path = os.getcwd()
        os.chdir(os.path.join(path, "lib"))
        subprocess.run(['python3', 'setup.py', 'build_ext'])
        os.chdir(ori_path)
    __file__ = pkg_resources.resource_filename(__name__, 'plugins.cpython-aarch64-linux-gnu.so')
    __loader__ = None; del __bootstrap__, __loader__
    spec = importlib.util.spec_from_file_location(__name__,__file__)
    mod = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(mod)
__bootstrap__()



import torch
import numpy as np
import tensorrt as trt

from .version_utils import (
    trt_version
)


def torch_dtype_to_trt(dtype):
    if trt_version() >= '7.0' and dtype == torch.bool:
        return trt.bool
    elif dtype == torch.int8:
        return trt.int8
    elif dtype == torch.int32:
        return trt.int32
    elif dtype == torch.float16:
        return trt.float16
    elif dtype == torch.float32:
        return trt.float32
    else:
        raise TypeError("%s is not supported by tensorrt" % dtype)


def torch_dtype_from_trt(dtype):
    if dtype == trt.int8:
        return torch.int8
    elif trt_version() >= '7.0' and dtype == trt.bool:
        return torch.bool
    elif dtype == trt.int32:
        return torch.int32
    elif dtype == trt.float16:
        return torch.float16
    elif dtype == trt.float32:
        return torch.float32
    else:
        raise TypeError("%s is not supported by torch" % dtype)


def torch_device_to_trt(device):
    if device.type == torch.device("cuda").type:
        return trt.TensorLocation.DEVICE
    elif device.type == torch.device("cpu").type:
        return trt.TensorLocation.HOST
    else:
        return TypeError("%s is not supported by tensorrt" % device)


def torch_device_from_trt(device):
    if device == trt.TensorLocation.DEVICE:
        return torch.device("cuda")
    elif device == trt.TensorLocation.HOST:
        return torch.device("cpu")
    else:
        return TypeError("%s is not supported by torch" % device)


def trt_int_dtype():
    if trt_version() >= "10.0":
        return np.int64
    else:
        return np.int32
    



import cv2, ctypes
import PIL.Image
import os, sys
import subprocess
import logging
import json
import time
import numpy as np
import ipywidgets.widgets as widgets
from IPython.display import display
import __main__


logging.basicConfig(
    format='%(asctime)s - [%(levelname)s] %(message)s',
    level=logging.INFO,
    datefmt='%m/%d/%Y %I:%M:%S %p',
)
isnotebook = __main__.isnotebook

def bgr8_to_jpeg(value):
    return bytes(cv2.imencode(".jpg", value)[1])

def import_tensorflow():
    global tf, K
    import tensorflow as tf
    import tensorflow.keras as __module
    import keras.backend as K
    
    for k in [m for m in dir(__module) if not "__" in m]:
        globals()[k]=__module.__dict__[k]

    import tensorflow.keras.models as __module
    
    for k in [m for m in dir(__module) if not "__" in m]:
        globals()[k]=__module.__dict__[k]

    import tensorflow.keras.losses as __module
    
    for k in [m for m in dir(__module) if not "__" in m]:
        globals()[k]=__module.__dict__[k]

    import tensorflow.keras.layers as __module
    
    for k in [m for m in dir(__module) if not "__" in m]:
        globals()[k]=__module.__dict__[k]

    import tensorflow.keras.optimizers as __module
    
    for k in [m for m in dir(__module) if not "__" in m]:
        globals()[k]=__module.__dict__[k]
        
    import tensorflow.keras.callbacks as __module
    
    for k in [m for m in dir(__module) if not "__" in m]:
        globals()[k]=__module.__dict__[k]

    import tensorflow.keras.utils as __module
    
    for k in [m for m in dir(__module) if not "__" in m]:
        globals()[k]=__module.__dict__[k]

    gpus=tf.config.experimental.list_physical_devices('GPU')

    if gpus:
        try:
            tf.config.set_visible_devices(gpus[0], 'GPU')
        except RuntimeError as e:
            logging.error(e)
            
def scheduler(epoch, learningrate):
    if epoch < 100:
        return learningrate
    else:
        return learningrate * tf.math.exp(-0.1)

class Object_Follow():
    list_path = os.path.abspath(__file__ + "/../lists/object.names")
    with open(list_path) as f:
        lines = f.readlines()
        label_list = [line.rstrip('\n') for line in lines]

    def __init__(self, camera=None, label_list=None):
        global trt, cuda
        import tensorrt as trt
        import pycuda.driver as cuda
        import pycuda.autoinit
        
        self.camera = camera
        self.default_path = os.path.abspath(__file__ + "/../models/yolov4-tiny/")

        if label_list != None:
            self.label_list = label_list

        if isnotebook:
            if camera is not None:
                self.probWidget = widgets.Image(
                    format="jpeg", width=camera.width, height=camera.height
                )
            else:
                self.probWidget = widgets.Image(
                    format="jpeg", width=300, height=300
                )
        else:
            self.probWidget = None
        
        self.cls_dict = {i: n for i, n in enumerate(self.label_list)}

    def __allocate_buffers(self, engine):
        inputs = []
        outputs = []
        bindings = []
        output_idx = 0
        stream = cuda.Stream()

        class HostDeviceMem(object):
            def __init__(self, host_mem, device_mem):
                self.host = host_mem
                self.device = device_mem

            def __repr__(self):
                return "Host:\n" + str(self.host) + "\nDevice:\n" + str(self.device)

        for binding in engine:
            binding_dims = engine.get_binding_shape(binding)
            if len(binding_dims) == 4:
                size = trt.volume(binding_dims)
            elif len(binding_dims) == 3:
                size = trt.volume(binding_dims) * engine.max_batch_size
            else:
                raise ValueError(
                    "bad dims of binding %s: %s" % (binding, str(binding_dims))
                )
            dtype = trt.nptype(engine.get_binding_dtype(binding))
            host_mem = cuda.pagelocked_empty(size, dtype)
            device_mem = cuda.mem_alloc(host_mem.nbytes)
            bindings.append(int(device_mem))
            if engine.binding_is_input(binding):
                inputs.append(HostDeviceMem(host_mem, device_mem))
            else:
                assert size % 7 == 0
                outputs.append(HostDeviceMem(host_mem, device_mem))
                output_idx += 1
        assert len(inputs) == 1
        assert len(outputs) == 1
        return inputs, outputs, bindings, stream

    def __preprocess_yolo(self, img, input_shape):
        img = cv2.resize(img, (input_shape[1], input_shape[0]))
        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        img = img.transpose((2, 0, 1)).astype(np.float32)
        img /= 255.0
        return img

    def __postprocess_yolo(
        self, trt_outputs, img_w, img_h, conf_th, nms_threshold, input_shape
    ):
        detections = []
        for o in trt_outputs:
            dets = o.reshape((-1, 7))
            dets = dets[dets[:, 4] * dets[:, 6] >= conf_th]
            detections.append(dets)
        detections = np.concatenate(detections, axis=0)

        if len(detections) == 0:
            boxes = np.zeros((0, 4), dtype=np.int)
            scores = np.zeros((0,), dtype=np.float32)
            classes = np.zeros((0,), dtype=np.float32)
        else:
            old_h, old_w = img_h, img_w
            detections[:, 0:4] *= np.array(
                [old_w, old_h, old_w, old_h], dtype=np.float32
            )

            nms_detections = np.zeros((0, 7), dtype=detections.dtype)
            for class_id in set(detections[:, 5]):
                idxs = np.where(detections[:, 5] == class_id)
                cls_detections = detections[idxs]

                x_coord = cls_detections[:, 0]
                y_coord = cls_detections[:, 1]
                width = cls_detections[:, 2]
                height = cls_detections[:, 3]
                box_confidences = cls_detections[:, 4] * cls_detections[:, 6]

                areas = width * height
                ordered = box_confidences.argsort()[::-1]

                keep = list()
                while ordered.size > 0:
                    i = ordered[0]
                    keep.append(i)
                    xx1 = np.maximum(x_coord[i], x_coord[ordered[1:]])
                    yy1 = np.maximum(y_coord[i], y_coord[ordered[1:]])
                    xx2 = np.minimum(
                        x_coord[i] + width[i], x_coord[ordered[1:]] + width[ordered[1:]]
                    )
                    yy2 = np.minimum(
                        y_coord[i] + height[i],
                        y_coord[ordered[1:]] + height[ordered[1:]],
                    )

                    width1 = np.maximum(0.0, xx2 - xx1 + 1)
                    height1 = np.maximum(0.0, yy2 - yy1 + 1)
                    intersection = width1 * height1
                    union = areas[i] + areas[ordered[1:]] - intersection
                    iou = intersection / union
                    indexes = np.where(iou <= nms_threshold)[0]
                    ordered = ordered[indexes + 1]

                keep = np.array(keep)
                nms_detections = np.concatenate(
                    [nms_detections, cls_detections[keep]], axis=0
                )

            xx = nms_detections[:, 0].reshape(-1, 1)
            yy = nms_detections[:, 1].reshape(-1, 1)
            ww = nms_detections[:, 2].reshape(-1, 1)
            hh = nms_detections[:, 3].reshape(-1, 1)
            boxes = np.concatenate([xx, yy, xx + ww, yy + hh], axis=1) + 0.5
            boxes = boxes.astype(np.int)
            scores = nms_detections[:, 4] * nms_detections[:, 6]
            classes = nms_detections[:, 5]
        return boxes, scores, classes

    def __inference_fn(self, context, bindings, inputs, outputs, stream):
        [cuda.memcpy_htod_async(inp.device, inp.host, stream) for inp in inputs]
        context.execute_async_v2(bindings=bindings, stream_handle=stream.handle)
        [cuda.memcpy_dtoh_async(out.host, out.device, stream) for out in outputs]
        stream.synchronize()
        return [out.host for out in outputs]

    def __draw_bboxes(self, img, boxes, clss, confs, idx, mx):
        for i, (bb, cl, cf) in enumerate(zip(boxes, clss, confs)):
            cl = int(cl)
            # if (cl != 12) and (cl != 13):
            x_min, y_min, x_max, y_max = bb[0], bb[1], bb[2], bb[3]
            if cl == idx and i == mx:
                box_color = (0, 255, 0)
                text_color = (0, 0, 0)
            else:
                box_color = (255, 0, 0)
                text_color = (255, 255, 255)
            cv2.rectangle(img, (x_min, y_min), (x_max, y_max), box_color, 2)
            txt_loc = (max(x_min + 2, 0), max(y_min + 2, 0))
            cls_name = self.cls_dict.get(cl, "CLS{}".format(cl))
            txt = "{} {:.2f}".format(cls_name, cf)
            img_h, img_w, _ = img.shape
            if txt_loc[0] >= img_w or txt_loc[1] >= img_h:
                return img
            margin = 3
            size = cv2.getTextSize(txt, cv2.FONT_HERSHEY_SIMPLEX, 0.5, 1)
            w = size[0][0] + margin * 2
            h = size[0][1] + margin * 2
            cv2.rectangle(
                img, (x_min - 1, y_min - 1 - h), (x_min + w, y_min), box_color, -1
            )
            cv2.putText(
                img,
                txt,
                (x_min + margin, y_min - margin - 2),
                cv2.FONT_HERSHEY_SIMPLEX,
                0.5,
                text_color,
                lineType=cv2.LINE_AA,
            )
        return img
    
    def generate_trt(self, model_path=None):
        self.layer_path = os.path.join(self.default_path, "libyolo_layer.so")
        if not os.path.exists(self.layer_path):
            ori_dir = os.getcwd()
            os.chdir(os.path.join(self.default_path, "plugins"))
            subprocess.run(['make'])
            os.chdir(ori_dir)
            
        from .converter import yolo_to_onnx
        from .converter import onnx_to_trt
        
        if model_path is None:
            import gdown
            
            ori_dir = os.getcwd()
            os.chdir(self.default_path)
            gdown.download('https://raw.githubusercontent.com/AlexeyAB/darknet/master/cfg/yolov4-tiny.cfg', 'yolov4-tiny.cfg', quiet=False)
            gdown.download('https://github.com/AlexeyAB/darknet/releases/download/darknet_yolo_v4_pre/yolov4-tiny.weights', 'yolov4-tiny.weights', quiet=False)
            yolo_to_onnx.convert('yolov4-tiny')
            onnx_to_trt.convert('yolov4-tiny')
            os.chdir(ori_dir)
        else:
            yolo_to_onnx.convert(model_path)
            onnx_to_trt.convert(model_path)

    def load_model(self, path=None):
        self.layer_path = os.path.join(self.default_path, "libyolo_layer.so")
        if not os.path.exists(self.layer_path):
            ori_dir = os.getcwd()
            os.chdir(os.path.join(self.default_path, "plugins"))
            subprocess.run(['make'])
            os.chdir(ori_dir)
        try:
            ctypes.cdll.LoadLibrary(self.layer_path)
        except OSError as e:
            raise SystemExit("ERROR: failed to load layer file.") from e
        
        if path is None:
            self.model_path = os.path.join(self.default_path, "yolov4-tiny.trt")
            if not os.path.exists(self.model_path):
                logging.info("YOLO trt model does not exist. Please, run generate_trt() first.")
        else:
            self.model_path = path

        self.trt_logger = trt.Logger(trt.Logger.INFO)
        with open(self.model_path, "rb") as f, trt.Runtime(self.trt_logger) as runtime:
            self.engine = runtime.deserialize_cuda_engine(f.read())
        if not self.engine:
            raise ValueError("failed to deserialize the engine.")

        binding = self.engine[0]
        binding_dims = self.engine.get_binding_shape(binding)
        try:
            self.input_shape = tuple(binding_dims[2:])
        except:
            raise ValueError(
                "bad dims of binding %s: %s" % (binding, str(binding_dims))
            )

        try:
            self.context = self.engine.create_execution_context()
            self.inputs, self.outputs, self.bindings, self.stream = (
                self.__allocate_buffers(self.engine)
            )
        except Exception as e:
            raise RuntimeError("fail to allocate CUDA resources") from e

    def show(self):
        display(self.probWidget)

    def detect(self, image=None, index=None, threshold=0.5, show=True, callback=None):
        if image is None:
            if self.camera is not None:
                image = self.camera.value
            else:
                raise logging.error("No Camera is available")

        width = image.shape[1]
        height = image.shape[0]

        if type(index) == str:
            try:
                index = self.label_list.index(index)
            except ValueError:
                logging.error("Index is not available.")
                return

        img_resized = self.__preprocess_yolo(image, self.input_shape)
        self.inputs[0].host = np.ascontiguousarray(img_resized)

        self.trt_outputs = self.__inference_fn(
            self.context, self.bindings, self.inputs, self.outputs, self.stream
        )
        boxes, scores, classes = self.__postprocess_yolo(
            self.trt_outputs,
            width,
            height,
            threshold,
            nms_threshold=0.5,
            input_shape=self.input_shape,
        )
        boxes[:, [0, 2]] = np.clip(boxes[:, [0, 2]], 0, width - 1)
        boxes[:, [1, 3]] = np.clip(boxes[:, [1, 3]], 0, height - 1)

        detections = []
        raw_value = []
        for bb, cl in zip(boxes, classes):
            detections.append(
                {
                    "label": int(cl),
                    "bbox": [
                        round(bb[0] / (width / 2) - 1, 2),
                        round(bb[1] / (height / 2) - 1, 2),
                        round(bb[2] / (width / 2) - 1, 2),
                        round(bb[3] / (height / 2) - 1, 2),
                    ],
                    "x": round(((bb[0] + bb[2]) / 2) / (width / 2) - 1, 5),
                    "y": round(((bb[1] + bb[3]) / 2) / (height / 2) - 1, 5),
                    "size_rate": round(
                        (bb[2] - bb[0]) * (bb[3] - bb[1]) / (width * height), 5
                    ),
                }
            )
            raw_value.append([int(cl), bb.tolist()])

        max_size = None
        if index is None:
            result = []
            for det in detections:
                result.append({"label": det["label"], "bbox": det["bbox"]})
        else:
            matching_detections = [d for d in detections if d["label"] == index]

            sizes = [
                det["size_rate"] for det in matching_detections if det["label"] == index
            ]
            if len(sizes) > 0:
                max_size = np.array(sizes).argmax()
                result = matching_detections[max_size]
                result.pop("label")
                result.pop("bbox")
            else:
                result = None

        if show:
            image = self.__draw_bboxes(image, boxes, classes, scores, index, max_size)
            if isnotebook:
                self.probWidget.value = bgr8_to_jpeg(image)
            self.value = image
        self.raw_value = raw_value

        return result

class Track_Follow():
    def __init__(self,camera=None):
        import_tensorflow()
        self.camera = camera
        self.default_path = os.path.abspath(__file__ + "/../models/Track_Follow/Track_Follow.h5")
        self.model = None
        if isnotebook:
            if camera is not None:
                self.probWidget = widgets.Image(
                    format="jpeg", width=camera.width, height=camera.height
                )
            else:
                self.probWidget = widgets.Image(
                    format="jpeg", width=300, height=300
                )
        else:
            self.probWidget = None

    def __load_layers(self):        
        def mish(x):
            return x * K.tanh(K.softplus(x))

        input1 = Input(shape=(150, 300, 3,))
        conv1 = Conv2D(filters=16, kernel_size=(3, 3), strides=(2, 2), padding="same")(input1)
        norm1 = BatchNormalization()(conv1)
        act1 = Activation(mish)(norm1)
        pool1 = MaxPooling2D(pool_size=(3, 3), strides=(2, 2))(act1)
        conv2 = Conv2D(filters=32, kernel_size=(3, 3), strides=(2, 2), padding="same")(pool1)
        norm2 = BatchNormalization()(conv2)
        act2 = Activation(mish)(norm2)
        conv3 = Conv2D(filters=32, kernel_size=(3, 3), strides=(1, 1), padding="same")(act2)
        norm3 = BatchNormalization()(conv3)
        act3 = Activation(mish)(norm3)
        add1 = Add()([act2, act3])
        conv4 = Conv2D(filters=64, kernel_size=(3, 3), strides=(2, 2), padding="same")(add1)
        norm4 = BatchNormalization()(conv4)
        act4 = Activation(mish)(norm4)
        conv5 = Conv2D(filters=64, kernel_size=(3, 3), strides=(1, 1), padding="same")(act4)
        norm5 = BatchNormalization()(conv5)
        act5 = Activation(mish)(norm5)
        add2 = Add()([act4, act5])
        conv6 = Conv2D(filters=128, kernel_size=(3, 3), strides=(2, 2), padding="same")(add2)
        norm6 = BatchNormalization()(conv6)
        act6 = Activation(mish)(norm6)
        conv7 = Conv2D(filters=128, kernel_size=(3, 3), strides=(1, 1), padding="same")(act6)
        norm7 = BatchNormalization()(conv7)
        act7 = Activation(mish)(norm7)
        add3 = Add()([act6, act7])
        conv8 = Conv2D(filters=256, kernel_size=(3, 3), strides=(2, 2), padding="same")(add3)
        norm8 = BatchNormalization()(conv8)
        act8 = Activation(mish)(norm8)
        conv9 = Conv2D(filters=512, kernel_size=(3, 3), strides=(2, 2), padding="same")(act8)
        norm9 = BatchNormalization()(conv9)
        act9 = Activation(mish)(norm9)
        flat1 = Flatten()(act9)
        dense1 = Dense(128)(flat1)
        norm10 = BatchNormalization()(dense1)
        act10 = Activation(mish)(norm10)
        dense2 = Dense(64)(act10)
        norm11 = BatchNormalization()(dense2)
        act11 = Activation(mish)(norm11)
        dense3 = Dense(64)(act11)
        norm12 = BatchNormalization()(dense3)
        act12 = Activation(mish)(norm12)
        dense4 = Dense(2, activation="tanh")(act12)

        self.model = Model(inputs=input1, outputs=dense4)
        
    def load_model(self,path=None):
        if path is None:
            path = self.default_path
            if not os.path.exists(path):
                import gdown
                
                ori_dir = os.getcwd()
                os.chdir(os.path.dirname(path))
                gdown.download('https://drive.google.com/uc?id=1rWF2iF556ntPM3EzDSpwzaDP_CEnLIG0', "Track_Follow.h5", quiet=False)
                os.chdir(ori_dir)
        elif not os.path.exists(path):
            logging.warning(f"{path} doesn't exist.")
            return

        if self.model is None:            
            self.__load_layers()
            opt = Adam()
            self.model.compile(optimizer=opt, loss='MAE')            
            self.model.optimizer.lr.numpy()

        self.model.load_weights(path)

    def show(self):
        display(self.probWidget)

    def run(self, value=None, callback=None):
        if self.model is None :
            logging.warning("Please load a trained model with load_model() method.")
            return
        
        img = self.camera.value if value is None else value
        crop_img = img[130:280,:300]
        x, y = self.model.predict(np.array([crop_img]).astype(np.float32))[0]
        height, width, _ = crop_img.shape
        
        if isnotebook:
            cX = int(width * (x / 2.0 + 0.5))
            cX = max(0, min(cX, 300))
            cY = 150
            self.value = cv2.circle(img, (cX, cY), 6, (255, 0, 0), 2)
            self.probWidget.value = bgr8_to_jpeg(self.value)

        if callback is not None:
            callback(result)
            
        result={"x":x,"y":y}

        return result
    
class PoseNet():
    def __init__(self,camera=None):
        global torch, transforms
        import torch
        import torchvision.transforms as transforms
        from jetson.torch2trt.parse_objects import ParseObjects
        
        self.camera = camera
        self.default_path = os.path.abspath(__file__ + "/../models/trtpose")
        if isnotebook:
            if camera is not None:
                self.probWidget = widgets.Image(
                    format="jpeg", width=camera.width, height=camera.height
                )
            else:
                self.probWidget = widgets.Image(
                    format="jpeg", width=300, height=300
                )
        else:
            self.probWidget = None
        
        self.WIDTH = 224
        self.HEIGHT = 224
        self.X_compress = 300.0 / self.WIDTH * 1.0
        self.Y_compress = 300.0 / self.HEIGHT * 1.0
        self.color = (0, 255, 0)
        
        self.mean = torch.Tensor([0.485, 0.456, 0.406]).cuda()
        self.std = torch.Tensor([0.229, 0.224, 0.225]).cuda()
        self.device = torch.device('cuda')
        self.human_pose = None
        self.topology = self.__topology()
        self.parse_objects = ParseObjects(self.topology)
        
    def __topology(self):
        json_path = os.path.join(self.default_path, "human_pose.json")
        
        with open(json_path, 'r') as f:
            self.human_pose = json.load(f)
            
        return self.__coco2topology(self.human_pose)
        
    def __coco2topology(self, coco_category):
        skeleton = coco_category['skeleton']
        K = len(skeleton)
        topology = torch.zeros((K, 4)).int()
        for k in range(K):
            topology[k][0] = 2 * k
            topology[k][1] = 2 * k + 1
            topology[k][2] = skeleton[k][0] - 1
            topology[k][3] = skeleton[k][1] - 1
        return topology
    
    def __preprocess(self, image):
        image = cv2.resize(image, dsize=(self.WIDTH, self.HEIGHT), interpolation=cv2.INTER_AREA)
                
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        image = PIL.Image.fromarray(image)
        image = transforms.functional.to_tensor(image).to(self.device)
        image.sub_(self.mean[:, None, None]).div_(self.std[:, None, None])
        return image[None, ...]
    
    def __get_keypoint(self, humans, hnum, peaks):
        kpoint = []
        human = humans[0][hnum]
        C = human.shape[0]
        for j in range(C):
            k = int(human[j])
            if k >= 0:
                peak = peaks[0][j][k]   # peak[1]:width, peak[0]:height
                peak = (j, float(peak[0]), float(peak[1]))
                kpoint.append(peak)
            else:    
                peak = (j, None, None)
                kpoint.append(peak)
        return kpoint
    
    def torch2trt(self, path=None):
        if path is None:
            path = os.path.join(self.default_path, "resnet18_baseline_att_224x224_A_epoch_249.pth")
            
            if not os.path.exists(path):
                import gdown
                
                ori_dir = os.getcwd()
                os.chdir(self.default_path)
                gdown.download('https://drive.google.com/uc?id=1XYDdCUdiF2xxx4rznmLb62SdOUZuoNbd', "resnet18_baseline_att_224x224_A_epoch_249.pth", quiet=False)
                os.chdir(ori_dir)
        elif not os.path.exists(path):
            logging.warning(f"{path} doesn't exist.")
            return
            
        import jetson.torch2trt 
        from jetson.torch2trt import models
        
        num_parts = len(self.human_pose['keypoints'])
        num_links = len(self.human_pose['skeleton'])

        data = torch.zeros((1, 3, self.HEIGHT, self.WIDTH)).cuda()
        
        model = models.resnet18_baseline_att(num_parts, 2 * num_links).cuda().eval()
        result = model.load_state_dict(torch.load(path))
        logging.info(f"{result}")
        logging.info("Converting the model now. This may take a while")
        self.model = jetson.torch2trt.torch2trt(model, [data], fp16_mode=True, max_workspace_size=1<<25)
        
        path = path.replace('.pth', '_trt.pth')
        torch.save(self.model.state_dict(), path)
        logging.info("Model is successfully converted")
        
    def load_model(self, path=None):
        if path is None:
            path = os.path.join(self.default_path, "resnet18_baseline_att_224x224_A_epoch_249_trt.pth")
            
            if not os.path.exists(path):
                logging.warning("There is no trt model. Please, run torch2trt() first.")
                return
        elif not os.path.exists(path):
            logging.warning(f"{path} doesn't exist.")
            return

        from jetson.torch2trt import trt_module

        self.model = trt_module.TRTModule()
        self.model.load_state_dict(torch.load(path))

    def show(self):
        display(self.probWidget)

    def run(self, value=None, callback=None):
        if self.model is None :
            logging.warning("Please load a trained model using load_model() method.")
            return
        
        img = self.camera.value if value is None else value
        frame = img.copy()
        data = self.__preprocess(frame)
        cmap, paf = self.model(data)
        cmap, paf = cmap.detach().cpu(), paf.detach().cpu()
        counts, objects, peaks = self.parse_objects(cmap, paf)
        
        for i in range(counts[0]):
            keypoints = self.__get_keypoint(objects, i, peaks)
            for j in range(len(keypoints)):
                if keypoints[j][1]:
                    x = round(keypoints[j][2] * self.WIDTH * self.X_compress)
                    y = round(keypoints[j][1] * self.HEIGHT * self.Y_compress)
                    cv2.circle(frame, (x, y), 3, self.color, 2)
                    cv2.putText(frame , "%d" % int(keypoints[j][0]), (x + 5, y),  cv2.FONT_HERSHEY_SIMPLEX, 1, (0, 255, 255), 1)
                    cv2.circle(frame, (x, y), 3, self.color, 2)
                    
        if isnotebook:
            self.probWidget.value = bgr8_to_jpeg(frame)

        return frame


import os

import numpy as np
import cv2
import pycuda.autoinit
import pycuda.driver as cuda
import tensorrt as trt


def _preprocess_yolo(img, input_shape):
    """Preprocess an image before TRT YOLO inferencing.

    # Args
        img: uint8 numpy array of shape either (img_h, img_w, 3)
             or (img_h, img_w)
        input_shape: a tuple of (H, W)

    # Returns
        preprocessed img: float32 numpy array of shape (3, H, W)
    """
    if img.ndim == 2:
        img = cv2.cvtColor(img, cv2.COLOR_GRAY2BGR)
    img = cv2.resize(img, (input_shape[1], input_shape[0]))
    img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
    img = img.transpose((2, 0, 1)).astype(np.float32)
    img /= 255.0
    return img


class YOLOEntropyCalibrator(trt.IInt8EntropyCalibrator2):
    """YOLOEntropyCalibrator

    This class implements TensorRT's IInt8EntropyCalibtrator2 interface.
    It reads all images from the specified directory and generates INT8
    calibration data for YOLO models accordingly.
    """

    def __init__(self, img_dir, net_hw, cache_file, batch_size=1):
        if not os.path.isdir(img_dir):
            raise FileNotFoundError('%s does not exist' % img_dir)
        if len(net_hw) != 2 or net_hw[0] % 32 or net_hw[1] % 32:
            raise ValueError('bad net shape: %s' % str(net_hw))

        super().__init__()  # trt.IInt8EntropyCalibrator2.__init__(self)

        self.img_dir = img_dir
        self.net_hw = net_hw
        self.cache_file = cache_file
        self.batch_size = batch_size
        self.blob_size = 3 * net_hw[0] * net_hw[1] * np.dtype('float32').itemsize * batch_size

        self.jpgs = [f for f in os.listdir(img_dir) if f.endswith('.jpg')]
        # The number "500" is NVIDIA's suggestion.  See here:
        # https://docs.nvidia.com/deeplearning/tensorrt/developer-guide/index.html#optimizing_int8_c
        if len(self.jpgs) < 500:
            print('WARNING: found less than 500 images in %s!' % img_dir)
        self.current_index = 0

        # Allocate enough memory for a whole batch.
        self.device_input = cuda.mem_alloc(self.blob_size)

    def __del__(self):
        del self.device_input  # free CUDA memory

    def get_batch_size(self):
        return self.batch_size

    def get_batch(self, names):
        if self.current_index + self.batch_size > len(self.jpgs):
            return None
        current_batch = int(self.current_index / self.batch_size)

        batch = []
        for i in range(self.batch_size):
            img_path = os.path.join(
                self.img_dir, self.jpgs[self.current_index + i])
            img = cv2.imread(img_path)
            assert img is not None, 'failed to read %s' % img_path
            batch.append(_preprocess_yolo(img, self.net_hw))
        batch = np.stack(batch)
        assert batch.nbytes == self.blob_size

        cuda.memcpy_htod(self.device_input, np.ascontiguousarray(batch))
        self.current_index += self.batch_size
        return [self.device_input]

    def read_calibration_cache(self):
        # If there is a cache, use it instead of calibrating again.
        # Otherwise, implicitly return None.
        if os.path.exists(self.cache_file):
            with open(self.cache_file, 'rb') as f:
                return f.read()

    def write_calibration_cache(self, cache):
        with open(self.cache_file, 'wb') as f:
            f.write(cache)



import torch
import torchvision
from .common import *


class DenseNetBackbone(torch.nn.Module):
    
    def __init__(self, densenet):
        super(DenseNetBackbone, self).__init__()
        self.densenet = densenet
    
    def forward(self, x):
        x = self.densenet.features(x)
        return x
    
    
def _densenet_pose(cmap_channels, paf_channels, upsample_channels, densenet, feature_channels, num_upsample, num_flat):
    model = torch.nn.Sequential(
        DenseNetBackbone(densenet),
        CmapPafHead(feature_channels, cmap_channels, paf_channels, upsample_channels, num_upsample=num_upsample, num_flat=num_flat)
    )
    return model
    
    
def _densenet_pose_att(cmap_channels, paf_channels, upsample_channels, densenet, feature_channels, num_upsample, num_flat):
    model = torch.nn.Sequential(
        DenseNetBackbone(densenet),
        CmapPafHeadAttention(feature_channels, cmap_channels, paf_channels, upsample_channels, num_upsample=num_upsample, num_flat=num_flat)
    )
    return model

    
def densenet121_baseline(cmap_channels, paf_channels, upsample_channels=256, pretrained=True, num_upsample=3, num_flat=0):
    densenet = torchvision.models.densenet121(pretrained=pretrained)
    return _densenet_pose(cmap_channels, paf_channels, upsample_channels, densenet, 1024, num_upsample, num_flat)


def densenet169_baseline(cmap_channels, paf_channels, upsample_channels=256, pretrained=True, num_upsample=3, num_flat=0):
    densenet = torchvision.models.densenet169(pretrained=pretrained)
    return _densenet_pose(cmap_channels, paf_channels, upsample_channels, densenet, 1664, num_upsample, num_flat)


def densenet201_baseline(cmap_channels, paf_channels, upsample_channels=256, pretrained=True, num_upsample=3, num_flat=0):
    densenet = torchvision.models.densenet201(pretrained=pretrained)
    return _densenet_pose(cmap_channels, paf_channels, upsample_channels, densenet, 1920, num_upsample, num_flat)


def densenet161_baseline(cmap_channels, paf_channels, upsample_channels=256, pretrained=True, num_upsample=3, num_flat=0):
    densenet = torchvision.models.densenet161(pretrained=pretrained)
    return _densenet_pose(cmap_channels, paf_channels, upsample_channels, densenet, 2208, num_upsample, num_flat)


    
def densenet121_baseline_att(cmap_channels, paf_channels, upsample_channels=256, pretrained=True, num_upsample=3, num_flat=0):
    densenet = torchvision.models.densenet121(pretrained=pretrained)
    return _densenet_pose_att(cmap_channels, paf_channels, upsample_channels, densenet, 1024, num_upsample, num_flat)


def densenet169_baseline_att(cmap_channels, paf_channels, upsample_channels=256, pretrained=True, num_upsample=3, num_flat=0):
    densenet = torchvision.models.densenet169(pretrained=pretrained)
    return _densenet_pose_att(cmap_channels, paf_channels, upsample_channels, densenet, 1664, num_upsample, num_flat)


def densenet201_baseline_att(cmap_channels, paf_channels, upsample_channels=256, pretrained=True, num_upsample=3, num_flat=0):
    densenet = torchvision.models.densenet201(pretrained=pretrained)
    return _densenet_pose_att(cmap_channels, paf_channels, upsample_channels, densenet, 1920, num_upsample, num_flat)


def densenet161_baseline_att(cmap_channels, paf_channels, upsample_channels=256, pretrained=True, num_upsample=3, num_flat=0):
    densenet = torchvision.models.densenet161(pretrained=pretrained)
    return _densenet_pose_att(cmap_channels, paf_channels, upsample_channels, densenet, 2208, num_upsample, num_flat)





import torch
import torch.nn as nn
from jetson.torch2trt.torch2trt import *
import numpy as np
import ctypes


try:
    ctypes.CDLL('libtorch2trt_plugins.so')

    def create_reflection_pad_2d_plugin(paddingLeft, paddingRight, paddingTop, paddingBottom):

        registry = trt.get_plugin_registry()
        creator = registry.get_plugin_creator('ReflectionPad2dPlugin', '1', '')

        fc = trt.PluginFieldCollection([
            trt.PluginField(
                'paddingLeft',
                np.array([paddingLeft]).astype(np.int32),
                trt.PluginFieldType.INT32
            ),
            trt.PluginField(
                'paddingRight',
                np.array([paddingRight]).astype(np.int32),
                trt.PluginFieldType.INT32
            ),
            trt.PluginField(
                'paddingTop',
                np.array([paddingTop]).astype(np.int32),
                trt.PluginFieldType.INT32
            ),
            trt.PluginField(
                'paddingBottom',
                np.array([paddingBottom]).astype(np.int32),
                trt.PluginFieldType.INT32
            )
        ])

        return creator.create_plugin('', fc)
    @tensorrt_converter(nn.ReflectionPad2d.forward)
    def convert_reflection_pad(ctx):
        module = get_arg(ctx, 'self', pos=0, default=None)
        input = get_arg(ctx, 'x', pos=1, default=None)
        output = ctx.method_return
        input_trt = input._trt
        plugin = create_reflection_pad_2d_plugin(
            module.padding[0],
            module.padding[1],
            module.padding[2],
            module.padding[3]
        )
        layer = ctx.network.add_plugin_v2([input_trt], plugin)
        output._trt = layer.get_output(0)

except:
    pass


import torch
import tensorrt as trt
import os
from .flattener import Flattener

__all__ = [
    'DEFAULT_CALIBRATION_ALGORITHM',
    'DatasetCalibrator'
]


if trt.__version__ >= '5.1':
    DEFAULT_CALIBRATION_ALGORITHM = trt.CalibrationAlgoType.ENTROPY_CALIBRATION_2
else:
    DEFAULT_CALIBRATION_ALGORITHM = trt.CalibrationAlgoType.ENTROPY_CALIBRATION


class DatasetCalibrator(trt.IInt8Calibrator):

    def __init__(self, dataset, algorithm=DEFAULT_CALIBRATION_ALGORITHM, cache_file=None, flattener=None):
        super(DatasetCalibrator, self).__init__()
        self.dataset = dataset
        self.algorithm = algorithm
        self.count = 0
        self.cache_file = cache_file
        if flattener is None:
            flattener = Flattener.from_value(dataset[0])
        self.flattener = flattener

    def get_batch(self, *args, **kwargs):
        if self.count < len(self.dataset):
            tensors = self.flattener.flatten(self.dataset[self.count])
            bindings = [int(t.data_ptr()) for t in tensors]
            self.count += 1
            return bindings
        else:
            return []

    def get_algorithm(self):
        return self.algorithm

    def get_batch_size(self):
        return 1

    def read_calibration_cache(self, *args, **kwargs):
        if (self.cache_file is not None) and os.path.exists(self.cache_file):
            with open(self.cache_file, 'rb') as f:
                return f.read()
                
    def write_calibration_cache(self, cache, *args, **kwargs):
        if self.cache_file is not None:
            with open(self.cache_file, 'wb') as f:
                f.write(cache)



import sys
import os
import torch
sys.path.append(os.path.join(os.path.dirname(__file__), 'dla'))
import dla_up


class DlaWrapper(torch.nn.Module):
    def __init__(self, dla_fn, cmap_channels, paf_channels):
        super(DlaWrapper, self).__init__()
        self.backbone = dla_fn(cmap_channels + paf_channels, pretrained_base='imagenet')
        self.cmap_channels = cmap_channels
        self.paf_channels = paf_channels
        
    def forward(self, x):
        x = self.backbone(x)
        cmap, paf = torch.split(x, [self.cmap_channels, self.paf_channels], dim=1)
        return cmap, paf
    
    
def dla34up_pose(cmap_channels, paf_channels):
    return DlaWrapper(dla_up.dla34up, cmap_channels, paf_channels)

def dla60up_pose(cmap_channels, paf_channels):
    return DlaWrapper(dla_up.dla60up, cmap_channels, paf_channels)

def dla102up_pose(cmap_channels, paf_channels):
    return DlaWrapper(dla_up.dla102up, cmap_channels, paf_channels)

def dla169up_pose(cmap_channels, paf_channels):
    return DlaWrapper(dla_up.dla169up, cmap_channels, paf_channels)


import os
import sys
import argparse
from collections import OrderedDict

import numpy as np
import onnx
from onnx import helper, TensorProto


path = os.path.abspath(__file__+"/../..")        
sys.path.append(path)

MAX_BATCH_SIZE = 1


def rreplace(s, old, new, occurrence=1):
    """Replace old pattern in the string with new from the right."""
    return new.join(s.rsplit(old, occurrence))


def is_pan_arch(cfg_file_path):
    """Determine whether the yolo model is with PAN architecture."""
    with open(cfg_file_path, 'r') as f:
        cfg_lines = [l.strip() for l in f.readlines()]
    yolos_or_upsamples = [l for l in cfg_lines
                            if l in ['[yolo]', '[upsample]']]
    yolo_count = len([l for l in yolos_or_upsamples if l == '[yolo]'])
    upsample_count = len(yolos_or_upsamples) - yolo_count
    assert yolo_count in (2, 3, 4)  # at most 4 yolo layers
    assert upsample_count == yolo_count - 1 or upsample_count == 0
    # the model is with PAN if an upsample layer appears before the 1st yolo
    return yolos_or_upsamples[0] == '[upsample]'


def get_output_convs(layer_configs):
    """Find output conv layer names from layer configs.

    The output conv layers are those conv layers immediately proceeding
    the yolo layers.

    # Arguments
        layer_configs: output of the DarkNetParser, i.e. a OrderedDict of
                       the yolo layers.
    """
    output_convs = []
    previous_layer = None
    for current_layer in layer_configs.keys():
        if previous_layer is not None and current_layer.endswith('yolo'):
            assert previous_layer.endswith('convolutional')
            activation = layer_configs[previous_layer]['activation']
            if activation == 'linear':
                output_convs.append(previous_layer)
            elif activation == 'logistic':
                output_convs.append(previous_layer + '_lgx')
            else:
                raise TypeError('unexpected activation: %s' % activation)
        previous_layer = current_layer
    return output_convs


def get_category_num(cfg_file_path):
    """Find number of output classes of the yolo model."""
    with open(cfg_file_path, 'r') as f:
        cfg_lines = [l.strip() for l in f.readlines()]
    classes_lines = [l for l in cfg_lines if l.startswith('classes=')]
    assert len(set(classes_lines)) == 1
    return int(classes_lines[-1].split('=')[-1].strip())


def get_h_and_w(layer_configs):
    """Find input height and width of the yolo model from layer configs."""
    net_config = layer_configs['000_net']
    return net_config['height'], net_config['width']


def get_anchors(cfg_file_path):
    """Get anchors of all yolo layers from the cfg file."""
    with open(cfg_file_path, 'r') as f:
        cfg_lines = f.readlines()
    yolo_lines = [l.strip() for l in cfg_lines if l.startswith('[yolo]')]
    mask_lines = [l.strip() for l in cfg_lines if l.startswith('mask')]
    anch_lines = [l.strip() for l in cfg_lines if l.startswith('anchors')]
    assert len(mask_lines) == len(yolo_lines)
    assert len(anch_lines) == len(yolo_lines)
    anchor_list = eval('[%s]' % anch_lines[0].split('=')[-1])
    mask_strs = [l.split('=')[-1] for l in mask_lines]
    masks = [eval('[%s]' % s)  for s in mask_strs]
    anchors = []
    for mask in masks:
        curr_anchors = []
        for m in mask:
            curr_anchors.append(anchor_list[m * 2])
            curr_anchors.append(anchor_list[m * 2 + 1])
        anchors.append(curr_anchors)
    return anchors


def get_anchor_num(cfg_file_path):
    """Find number of anchors (masks) of the yolo model."""
    anchors = get_anchors(cfg_file_path)
    num_anchors = [len(a) // 2 for a in anchors]

    assert len(num_anchors) > 0, 'Found no `mask` fields in config'
    assert len(set(num_anchors)) == 1, 'Found different num anchors'

    return num_anchors[0]


class DarkNetParser(object):
    """Definition of a parser for DarkNet-based YOLO model."""

    def __init__(self, supported_layers=None):
        """Initializes a DarkNetParser object.

        Keyword argument:
        supported_layers -- a string list of supported layers in DarkNet naming convention,
        parameters are only added to the class dictionary if a parsed layer is included.
        """

        # A list of YOLO layers containing dictionaries with all layer
        # parameters:
        self.layer_configs = OrderedDict()
        self.supported_layers = supported_layers if supported_layers else \
                                ['net', 'convolutional', 'maxpool', 'shortcut',
                                 'route', 'upsample', 'yolo']
        self.layer_counter = 0

    def parse_cfg_file(self, cfg_file_path):
        """Takes the yolov?.cfg file and parses it layer by layer,
        appending each layer's parameters as a dictionary to layer_configs.

        Keyword argument:
        cfg_file_path
        """
        with open(cfg_file_path, 'r') as cfg_file:
            remainder = cfg_file.read()
            while remainder is not None:
                layer_dict, layer_name, remainder = self._next_layer(remainder)
                if layer_dict is not None:
                    self.layer_configs[layer_name] = layer_dict
        return self.layer_configs

    def _next_layer(self, remainder):
        """Takes in a string and segments it by looking for DarkNet delimiters.
        Returns the layer parameters and the remaining string after the last delimiter.
        Example for the first Conv layer in yolo.cfg ...

        [convolutional]
        batch_normalize=1
        filters=32
        size=3
        stride=1
        pad=1
        activation=leaky

        ... becomes the following layer_dict return value:
        {'activation': 'leaky', 'stride': 1, 'pad': 1, 'filters': 32,
        'batch_normalize': 1, 'type': 'convolutional', 'size': 3}.

        '001_convolutional' is returned as layer_name, and all lines that follow in yolo.cfg
        are returned as the next remainder.

        Keyword argument:
        remainder -- a string with all raw text after the previously parsed layer
        """
        remainder = remainder.split('[', 1)
        while len(remainder[0]) > 0 and remainder[0][-1] == '#':
            # '#[...' case (the left bracket is proceeded by a pound sign),
            # assuming this layer is commented out, so go find the next '['
            remainder = remainder[1].split('[', 1)
        if len(remainder) == 2:
            remainder = remainder[1]
        else:
            # no left bracket found in remainder
            return None, None, None
        remainder = remainder.split(']', 1)
        if len(remainder) == 2:
            layer_type, remainder = remainder
        else:
            # no right bracket
            raise ValueError('no closing bracket!')
        if layer_type not in self.supported_layers:
            raise ValueError('%s layer not supported!' % layer_type)

        out = remainder.split('\n[', 1)
        if len(out) == 2:
            layer_param_block, remainder = out[0], '[' + out[1]
        else:
            layer_param_block, remainder = out[0], ''
        layer_param_lines = layer_param_block.split('\n')
        # remove empty lines
        layer_param_lines = [l.lstrip() for l in layer_param_lines if l.lstrip()]
        # don't parse yolo layers
        if layer_type == 'yolo':  layer_param_lines = []
        skip_params = ['steps', 'scales'] if layer_type == 'net' else []
        layer_name = str(self.layer_counter).zfill(3) + '_' + layer_type
        layer_dict = dict(type=layer_type)
        for param_line in layer_param_lines:
            param_line = param_line.split('#')[0]
            if not param_line:  continue
            assert '[' not in param_line
            param_type, param_value = self._parse_params(param_line, skip_params)
            layer_dict[param_type] = param_value
        self.layer_counter += 1
        return layer_dict, layer_name, remainder

    def _parse_params(self, param_line, skip_params=None):
        """Identifies the parameters contained in one of the cfg file and returns
        them in the required format for each parameter type, e.g. as a list, an int or a float.

        Keyword argument:
        param_line -- one parsed line within a layer block
        """
        param_line = param_line.replace(' ', '')
        param_type, param_value_raw = param_line.split('=')
        assert param_value_raw
        param_value = None
        if skip_params and param_type in skip_params:
            param_type = None
        elif param_type == 'layers':
            layer_indexes = list()
            for index in param_value_raw.split(','):
                layer_indexes.append(int(index))
            param_value = layer_indexes
        elif isinstance(param_value_raw, str) and not param_value_raw.isalpha():
            condition_param_value_positive = param_value_raw.isdigit()
            condition_param_value_negative = param_value_raw[0] == '-' and \
                param_value_raw[1:].isdigit()
            if condition_param_value_positive or condition_param_value_negative:
                param_value = int(param_value_raw)
            else:
                param_value = float(param_value_raw)
        else:
            param_value = str(param_value_raw)
        return param_type, param_value


class MajorNodeSpecs(object):
    """Helper class used to store the names of ONNX output names,
    corresponding to the output of a DarkNet layer and its output channels.
    Some DarkNet layers are not created and there is no corresponding ONNX node,
    but we still need to track them in order to set up skip connections.
    """

    def __init__(self, name, channels):
        """ Initialize a MajorNodeSpecs object.

        Keyword arguments:
        name -- name of the ONNX node
        channels -- number of output channels of this node
        """
        self.name = name
        self.channels = channels
        self.created_onnx_node = False
        if name is not None and isinstance(channels, int) and channels > 0:
            self.created_onnx_node = True


class ConvParams(object):
    """Helper class to store the hyper parameters of a Conv layer,
    including its prefix name in the ONNX graph and the expected dimensions
    of weights for convolution, bias, and batch normalization.

    Additionally acts as a wrapper for generating safe names for all
    weights, checking on feasible combinations.
    """

    def __init__(self, node_name, batch_normalize, conv_weight_dims):
        """Constructor based on the base node name (e.g. 101_convolutional), the batch
        normalization setting, and the convolutional weights shape.

        Keyword arguments:
        node_name -- base name of this YOLO convolutional layer
        batch_normalize -- bool value if batch normalization is used
        conv_weight_dims -- the dimensions of this layer's convolutional weights
        """
        self.node_name = node_name
        self.batch_normalize = batch_normalize
        assert len(conv_weight_dims) == 4
        self.conv_weight_dims = conv_weight_dims

    def generate_param_name(self, param_category, suffix):
        """Generates a name based on two string inputs,
        and checks if the combination is valid."""
        assert suffix
        assert param_category in ['bn', 'conv']
        assert(suffix in ['scale', 'mean', 'var', 'weights', 'bias'])
        if param_category == 'bn':
            assert self.batch_normalize
            assert suffix in ['scale', 'bias', 'mean', 'var']
        elif param_category == 'conv':
            assert suffix in ['weights', 'bias']
            if suffix == 'bias':
                assert not self.batch_normalize
        param_name = self.node_name + '_' + param_category + '_' + suffix
        return param_name

class ResizeParams(object):
    #Helper class to store the scale parameter for an Resize node.

    def __init__(self, node_name, value):
        """Constructor based on the base node name (e.g. 86_Resize),
        and the value of the scale input tensor.

        Keyword arguments:
        node_name -- base name of this YOLO Resize layer
        value -- the value of the scale input to the Resize layer as numpy array
        """
        self.node_name = node_name
        self.value = value

    def generate_param_name(self):
        """Generates the scale parameter name for the Resize node."""
        param_name = self.node_name + '_' + "scale"
        return param_name

    def generate_roi_name(self):
        """Generates the roi input name for the Resize node."""
        param_name = self.node_name + '_' + "roi"
        return param_name

class WeightLoader(object):
    """Helper class used for loading the serialized weights of a binary file stream
    and returning the initializers and the input tensors required for populating
    the ONNX graph with weights.
    """

    def __init__(self, weights_file_path):
        """Initialized with a path to the YOLO .weights file.

        Keyword argument:
        weights_file_path -- path to the weights file.
        """
        self.weights_file = self._open_weights_file(weights_file_path)

    def load_resize_scales(self, resize_params):
        """Returns the initializers with the value of the scale input
        tensor given by resize_params.

        Keyword argument:
        resize_params -- a ResizeParams object
        """
        initializer = list()
        inputs = list()
        name = resize_params.generate_param_name()
        shape = resize_params.value.shape
        data = resize_params.value
        scale_init = helper.make_tensor(
            name, TensorProto.FLOAT, shape, data)
        scale_input = helper.make_tensor_value_info(
            name, TensorProto.FLOAT, shape)
        initializer.append(scale_init)
        inputs.append(scale_input)

        # In opset 11 an additional input named roi is required. Create a dummy tensor to satisfy this.
        # It is a 1D tensor of size of the rank of the input (4)
        rank = 4
        roi_name = resize_params.generate_roi_name()
        roi_input = helper.make_tensor_value_info(roi_name, TensorProto.FLOAT, [rank])
        roi_init = helper.make_tensor(roi_name, TensorProto.FLOAT, [rank], [0,0,0,0])
        initializer.append(roi_init)
        inputs.append(roi_input)

        return initializer, inputs

    def load_conv_weights(self, conv_params):
        """Returns the initializers with weights from the weights file and
        the input tensors of a convolutional layer for all corresponding ONNX nodes.

        Keyword argument:
        conv_params -- a ConvParams object
        """
        initializer = list()
        inputs = list()
        if conv_params.batch_normalize:
            bias_init, bias_input = self._create_param_tensors(
                conv_params, 'bn', 'bias')
            bn_scale_init, bn_scale_input = self._create_param_tensors(
                conv_params, 'bn', 'scale')
            bn_mean_init, bn_mean_input = self._create_param_tensors(
                conv_params, 'bn', 'mean')
            bn_var_init, bn_var_input = self._create_param_tensors(
                conv_params, 'bn', 'var')
            initializer.extend(
                [bn_scale_init, bias_init, bn_mean_init, bn_var_init])
            inputs.extend([bn_scale_input, bias_input,
                           bn_mean_input, bn_var_input])
        else:
            bias_init, bias_input = self._create_param_tensors(
                conv_params, 'conv', 'bias')
            initializer.append(bias_init)
            inputs.append(bias_input)
        conv_init, conv_input = self._create_param_tensors(
            conv_params, 'conv', 'weights')
        initializer.append(conv_init)
        inputs.append(conv_input)
        return initializer, inputs

    def _open_weights_file(self, weights_file_path):
        """Opens a YOLO DarkNet file stream and skips the header.

        Keyword argument:
        weights_file_path -- path to the weights file.
        """
        weights_file = open(weights_file_path, 'rb')
        length_header = 5
        np.ndarray(shape=(length_header, ), dtype='int32',
                   buffer=weights_file.read(length_header * 4))
        return weights_file

    def _create_param_tensors(self, conv_params, param_category, suffix):
        """Creates the initializers with weights from the weights file together with
        the input tensors.

        Keyword arguments:
        conv_params -- a ConvParams object
        param_category -- the category of parameters to be created ('bn' or 'conv')
        suffix -- a string determining the sub-type of above param_category (e.g.,
        'weights' or 'bias')
        """
        param_name, param_data, param_data_shape = self._load_one_param_type(
            conv_params, param_category, suffix)

        initializer_tensor = helper.make_tensor(
            param_name, TensorProto.FLOAT, param_data_shape, param_data)
        input_tensor = helper.make_tensor_value_info(
            param_name, TensorProto.FLOAT, param_data_shape)
        return initializer_tensor, input_tensor

    def _load_one_param_type(self, conv_params, param_category, suffix):
        """Deserializes the weights from a file stream in the DarkNet order.

        Keyword arguments:
        conv_params -- a ConvParams object
        param_category -- the category of parameters to be created ('bn' or 'conv')
        suffix -- a string determining the sub-type of above param_category (e.g.,
        'weights' or 'bias')
        """
        param_name = conv_params.generate_param_name(param_category, suffix)
        channels_out, channels_in, filter_h, filter_w = conv_params.conv_weight_dims
        if param_category == 'bn':
            param_shape = [channels_out]
        elif param_category == 'conv':
            if suffix == 'weights':
                param_shape = [channels_out, channels_in, filter_h, filter_w]
            elif suffix == 'bias':
                param_shape = [channels_out]
        param_size = np.product(np.array(param_shape))
        param_data = np.ndarray(
            shape=param_shape,
            dtype='float32',
            buffer=self.weights_file.read(param_size * 4))
        param_data = param_data.flatten().astype(float)
        return param_name, param_data, param_shape


class GraphBuilderONNX(object):
    """Class for creating an ONNX graph from a previously generated list of layer dictionaries."""

    def __init__(self, model_name, output_tensors, batch_size):
        """Initialize with all DarkNet default parameters used creating
        YOLO, and specify the output tensors as an OrderedDict for their
        output dimensions with their names as keys.

        Keyword argument:
        output_tensors -- the output tensors as an OrderedDict containing the keys'
        output dimensions
        """
        self.model_name = model_name
        self.output_tensors = output_tensors
        self._nodes = list()
        self.graph_def = None
        self.input_tensor = None
        self.epsilon_bn = 1e-5
        self.momentum_bn = 0.99
        self.alpha_lrelu = 0.1
        self.param_dict = OrderedDict()
        self.major_node_specs = list()
        self.batch_size = batch_size
        self.route_spec = 0  # keeping track of the current active 'route'

    def build_onnx_graph(
            self,
            layer_configs,
            weights_file_path,
            verbose=True):
        """Iterate over all layer configs (parsed from the DarkNet
        representation of YOLO), create an ONNX graph, populate it with
        weights from the weights file and return the graph definition.

        Keyword arguments:
        layer_configs -- an OrderedDict object with all parsed layers' configurations
        weights_file_path -- location of the weights file
        verbose -- toggles if the graph is printed after creation (default: True)
        """
        for layer_name in layer_configs.keys():
            layer_dict = layer_configs[layer_name]
            major_node_specs = self._make_onnx_node(layer_name, layer_dict)
            if major_node_specs.name is not None:
                self.major_node_specs.append(major_node_specs)
        # remove dummy 'route' and 'yolo' nodes
        self.major_node_specs = [node for node in self.major_node_specs
                                      if 'dummy' not in node.name]
        outputs = list()
        for tensor_name in self.output_tensors.keys():
            output_dims = [self.batch_size, ] + \
                self.output_tensors[tensor_name]
            output_tensor = helper.make_tensor_value_info(
                tensor_name, TensorProto.FLOAT, output_dims)
            outputs.append(output_tensor)
        inputs = [self.input_tensor]
        weight_loader = WeightLoader(weights_file_path)
        initializer = list()
        # If a layer has parameters, add them to the initializer and input lists.
        for layer_name in self.param_dict.keys():
            _, layer_type = layer_name.split('_', 1)
            params = self.param_dict[layer_name]
            if layer_type == 'convolutional':
                initializer_layer, inputs_layer = weight_loader.load_conv_weights(
                    params)
                initializer.extend(initializer_layer)
                inputs.extend(inputs_layer)
            elif layer_type == 'upsample':
                initializer_layer, inputs_layer = weight_loader.load_resize_scales(
                    params)
                initializer.extend(initializer_layer)
                inputs.extend(inputs_layer)
        del weight_loader
        self.graph_def = helper.make_graph(
            nodes=self._nodes,
            name=self.model_name,
            inputs=inputs,
            outputs=outputs,
            initializer=initializer
        )
        if verbose:
            print(helper.printable_graph(self.graph_def))
        model_def = helper.make_model(self.graph_def,
                                      producer_name='NVIDIA TensorRT sample')
        return model_def

    def _make_onnx_node(self, layer_name, layer_dict):
        """Take in a layer parameter dictionary, choose the correct function for
        creating an ONNX node and store the information important to graph creation
        as a MajorNodeSpec object.

        Keyword arguments:
        layer_name -- the layer's name (also the corresponding key in layer_configs)
        layer_dict -- a layer parameter dictionary (one element of layer_configs)
        """
        layer_type = layer_dict['type']
        if self.input_tensor is None:
            if layer_type == 'net':
                major_node_output_name, major_node_output_channels = self._make_input_tensor(
                    layer_name, layer_dict)
                major_node_specs = MajorNodeSpecs(major_node_output_name,
                                                  major_node_output_channels)
            else:
                raise ValueError('The first node has to be of type "net".')
        else:
            node_creators = dict()
            node_creators['convolutional'] = self._make_conv_node
            node_creators['maxpool'] = self._make_maxpool_node
            node_creators['shortcut'] = self._make_shortcut_node
            node_creators['route'] = self._make_route_node
            node_creators['upsample'] = self._make_resize_node
            node_creators['yolo'] = self._make_yolo_node

            if layer_type in node_creators.keys():
                major_node_output_name, major_node_output_channels = \
                    node_creators[layer_type](layer_name, layer_dict)
                major_node_specs = MajorNodeSpecs(major_node_output_name,
                                                  major_node_output_channels)
            else:
                raise TypeError('layer of type %s not supported' % layer_type)
        return major_node_specs

    def _make_input_tensor(self, layer_name, layer_dict):
        """Create an ONNX input tensor from a 'net' layer and store the batch size.

        Keyword arguments:
        layer_name -- the layer's name (also the corresponding key in layer_configs)
        layer_dict -- a layer parameter dictionary (one element of layer_configs)
        """
        #batch_size = layer_dict['batch']
        channels = layer_dict['channels']
        height = layer_dict['height']
        width = layer_dict['width']
        #self.batch_size = batch_size
        input_tensor = helper.make_tensor_value_info(
            str(layer_name), TensorProto.FLOAT, [
                self.batch_size, channels, height, width])
        self.input_tensor = input_tensor
        return layer_name, channels

    def _get_previous_node_specs(self, target_index=0):
        """Get a previously ONNX node.

        Target index can be passed for jumping to a specific index.

        Keyword arguments:
        target_index -- optional for jumping to a specific index,
                        default: 0 for the previous element, while
                        taking 'route' spec into account
        """
        if target_index == 0:
            if self.route_spec != 0:
                previous_node = self.major_node_specs[self.route_spec]
                assert 'dummy' not in previous_node.name
                self.route_spec = 0
            else:
                previous_node = self.major_node_specs[-1]
        else:
            previous_node = self.major_node_specs[target_index]
        assert previous_node.created_onnx_node
        return previous_node

    def _make_conv_node(self, layer_name, layer_dict):
        """Create an ONNX Conv node with optional batch normalization and
        activation nodes.

        Keyword arguments:
        layer_name -- the layer's name (also the corresponding key in layer_configs)
        layer_dict -- a layer parameter dictionary (one element of layer_configs)
        """
        previous_node_specs = self._get_previous_node_specs()
        inputs = [previous_node_specs.name]
        previous_channels = previous_node_specs.channels
        kernel_size = layer_dict['size']
        stride = layer_dict['stride']
        filters = layer_dict['filters']
        batch_normalize = False
        if layer_dict.get('batch_normalize', 0) > 0:
            batch_normalize = True

        kernel_shape = [kernel_size, kernel_size]
        weights_shape = [filters, previous_channels] + kernel_shape
        conv_params = ConvParams(layer_name, batch_normalize, weights_shape)

        strides = [stride, stride]
        dilations = [1, 1]
        weights_name = conv_params.generate_param_name('conv', 'weights')
        inputs.append(weights_name)
        if not batch_normalize:
            bias_name = conv_params.generate_param_name('conv', 'bias')
            inputs.append(bias_name)

        conv_node = helper.make_node(
            'Conv',
            inputs=inputs,
            outputs=[layer_name],
            kernel_shape=kernel_shape,
            strides=strides,
            auto_pad='SAME_LOWER',
            dilations=dilations,
            name=layer_name
        )
        self._nodes.append(conv_node)
        inputs = [layer_name]
        layer_name_output = layer_name

        if batch_normalize:
            layer_name_bn = layer_name + '_bn'
            bn_param_suffixes = ['scale', 'bias', 'mean', 'var']
            for suffix in bn_param_suffixes:
                bn_param_name = conv_params.generate_param_name('bn', suffix)
                inputs.append(bn_param_name)
            batchnorm_node = helper.make_node(
                'BatchNormalization',
                inputs=inputs,
                outputs=[layer_name_bn],
                epsilon=self.epsilon_bn,
                momentum=self.momentum_bn,
                name=layer_name_bn
            )
            self._nodes.append(batchnorm_node)
            inputs = [layer_name_bn]
            layer_name_output = layer_name_bn

        if layer_dict['activation'] == 'leaky':
            layer_name_lrelu = layer_name + '_lrelu'

            lrelu_node = helper.make_node(
                'LeakyRelu',
                inputs=inputs,
                outputs=[layer_name_lrelu],
                name=layer_name_lrelu,
                alpha=self.alpha_lrelu
            )
            self._nodes.append(lrelu_node)
            inputs = [layer_name_lrelu]
            layer_name_output = layer_name_lrelu
        elif layer_dict['activation'] == 'mish':
            layer_name_softplus = layer_name + '_softplus'
            layer_name_tanh = layer_name + '_tanh'
            layer_name_mish = layer_name + '_mish'

            softplus_node = helper.make_node(
                'Softplus',
                inputs=inputs,
                outputs=[layer_name_softplus],
                name=layer_name_softplus
            )
            self._nodes.append(softplus_node)
            tanh_node = helper.make_node(
                'Tanh',
                inputs=[layer_name_softplus],
                outputs=[layer_name_tanh],
                name=layer_name_tanh
            )
            self._nodes.append(tanh_node)

            inputs.append(layer_name_tanh)
            mish_node = helper.make_node(
                'Mul',
                inputs=inputs,
                outputs=[layer_name_mish],
                name=layer_name_mish
            )
            self._nodes.append(mish_node)

            inputs = [layer_name_mish]
            layer_name_output = layer_name_mish
        elif layer_dict['activation'] == 'swish':
            layer_name_sigmoid = layer_name + '_sigmoid'
            layer_name_swish = layer_name + '_swish'

            sigmoid_node = helper.make_node(
                'Sigmoid',
                inputs=inputs,
                outputs=[layer_name_sigmoid],
                name=layer_name_sigmoid
            )
            self._nodes.append(sigmoid_node)

            inputs.append(layer_name_sigmoid)
            swish_node = helper.make_node(
                'Mul',
                inputs=inputs,
                outputs=[layer_name_swish],
                name=layer_name_swish
            )
            self._nodes.append(swish_node)

            inputs = [layer_name_swish]
            layer_name_output = layer_name_swish
        elif layer_dict['activation'] == 'logistic':
            layer_name_lgx = layer_name + '_lgx'

            lgx_node = helper.make_node(
                'Sigmoid',
                inputs=inputs,
                outputs=[layer_name_lgx],
                name=layer_name_lgx
            )
            self._nodes.append(lgx_node)
            inputs = [layer_name_lgx]
            layer_name_output = layer_name_lgx
        elif layer_dict['activation'] == 'linear':
            pass
        else:
            raise TypeError('%s activation not supported' % layer_dict['activation'])

        self.param_dict[layer_name] = conv_params
        return layer_name_output, filters

    def _make_shortcut_node(self, layer_name, layer_dict):
        """Create an ONNX Add node with the shortcut properties from
        the DarkNet-based graph.

        Keyword arguments:
        layer_name -- the layer's name (also the corresponding key in layer_configs)
        layer_dict -- a layer parameter dictionary (one element of layer_configs)
        """
        shortcut_index = layer_dict['from']
        activation = layer_dict['activation']
        assert activation == 'linear'

        first_node_specs = self._get_previous_node_specs()
        second_node_specs = self._get_previous_node_specs(
            target_index=shortcut_index)
        assert first_node_specs.channels == second_node_specs.channels
        channels = first_node_specs.channels
        inputs = [first_node_specs.name, second_node_specs.name]
        shortcut_node = helper.make_node(
            'Add',
            inputs=inputs,
            outputs=[layer_name],
            name=layer_name,
        )
        self._nodes.append(shortcut_node)
        return layer_name, channels

    def _make_route_node(self, layer_name, layer_dict):
        """If the 'layers' parameter from the DarkNet configuration is only one index, continue
        node creation at the indicated (negative) index. Otherwise, create an ONNX Concat node
        with the route properties from the DarkNet-based graph.

        Keyword arguments:
        layer_name -- the layer's name (also the corresponding key in layer_configs)
        layer_dict -- a layer parameter dictionary (one element of layer_configs)
        """
        route_node_indexes = layer_dict['layers']
        if len(route_node_indexes) == 1:
            if 'groups' in layer_dict.keys():
                # for CSPNet-kind of architecture
                assert 'group_id' in layer_dict.keys()
                groups = layer_dict['groups']
                group_id = int(layer_dict['group_id'])
                assert group_id < groups
                index = route_node_indexes[0]
                if index > 0:
                    # +1 for input node (same reason as below)
                    index += 1
                route_node_specs = self._get_previous_node_specs(
                    target_index=index)
                assert route_node_specs.channels % groups == 0
                channels = route_node_specs.channels // groups

                outputs = [layer_name + '_dummy%d' % i for i in range(groups)]
                outputs[group_id] = layer_name
                route_node = helper.make_node(
                    'Split',
                    axis=1,
                    #split=[channels] * groups,  # not needed for opset 11
                    inputs=[route_node_specs.name],
                    outputs=outputs,
                    name=layer_name,
                )
                self._nodes.append(route_node)
            else:
                if route_node_indexes[0] < 0:
                    # route should skip self, thus -1
                    self.route_spec = route_node_indexes[0] - 1
                elif route_node_indexes[0] > 0:
                    # +1 for input node (same reason as below)
                    self.route_spec = route_node_indexes[0] + 1
                # This dummy route node would be removed in the end.
                layer_name = layer_name + '_dummy'
                channels = 1
        else:
            assert 'groups' not in layer_dict.keys(), \
                'groups not implemented for multiple-input route layer!'
            inputs = list()
            channels = 0
            for index in route_node_indexes:
                if index > 0:
                    # Increment by one because we count the input as
                    # a node (DarkNet does not)
                    index += 1
                route_node_specs = self._get_previous_node_specs(
                    target_index=index)
                inputs.append(route_node_specs.name)
                channels += route_node_specs.channels
            assert inputs
            assert channels > 0

            route_node = helper.make_node(
                'Concat',
                axis=1,
                inputs=inputs,
                outputs=[layer_name],
                name=layer_name,
            )
            self._nodes.append(route_node)
        return layer_name, channels

    def _make_resize_node(self, layer_name, layer_dict):
        """Create an ONNX Resize node with the properties from
        the DarkNet-based graph.

        Keyword arguments:
        layer_name -- the layer's name (also the corresponding key in layer_configs)
        layer_dict -- a layer parameter dictionary (one element of layer_configs)
        """
        resize_scale_factors = float(layer_dict['stride'])
        # Create the scale factor array with node parameters
        scales=np.array([1.0, 1.0, resize_scale_factors, resize_scale_factors]).astype(np.float32)
        previous_node_specs = self._get_previous_node_specs()
        inputs = [previous_node_specs.name]

        channels = previous_node_specs.channels
        assert channels > 0
        resize_params = ResizeParams(layer_name, scales)

        # roi input is the second input, so append it before scales
        roi_name = resize_params.generate_roi_name()
        inputs.append(roi_name)

        scales_name = resize_params.generate_param_name()
        inputs.append(scales_name)

        resize_node = helper.make_node(
            'Resize',
            coordinate_transformation_mode='asymmetric',
            mode='nearest',
            nearest_mode='floor',
            inputs=inputs,
            outputs=[layer_name],
            name=layer_name,
        )
        self._nodes.append(resize_node)
        self.param_dict[layer_name] = resize_params
        return layer_name, channels

    def _make_maxpool_node(self, layer_name, layer_dict):
        """Create an ONNX Maxpool node with the properties from
        the DarkNet-based graph.

        Keyword arguments:
        layer_name -- the layer's name (also the corresponding key in layer_configs)
        layer_dict -- a layer parameter dictionary (one element of layer_configs)
        """
        stride = layer_dict['stride']
        kernel_size = layer_dict['size']
        previous_node_specs = self._get_previous_node_specs()
        inputs = [previous_node_specs.name]
        channels = previous_node_specs.channels
        kernel_shape = [kernel_size, kernel_size]
        strides = [stride, stride]
        assert channels > 0
        maxpool_node = helper.make_node(
            'MaxPool',
            inputs=inputs,
            outputs=[layer_name],
            kernel_shape=kernel_shape,
            strides=strides,
            auto_pad='SAME_UPPER',
            name=layer_name,
        )
        self._nodes.append(maxpool_node)
        return layer_name, channels

    def _make_yolo_node(self, layer_name, layer_dict):
        """Create an ONNX Yolo node.

        These are dummy nodes which would be removed in the end.
        """
        channels = 1
        return layer_name + '_dummy', channels


def convert(model):
    if sys.version_info[0] < 3:
        raise SystemExit('ERROR: This modified version of yolov3_to_onnx.py '
                         'script is only compatible with python3...')

    cfg_file_path = '%s.cfg' % model
    if not os.path.isfile(cfg_file_path):
        raise SystemExit('ERROR: file (%s) not found!' % cfg_file_path)
    weights_file_path = '%s.weights' % model
    if not os.path.isfile(weights_file_path):
        raise SystemExit('ERROR: file (%s) not found!' % weights_file_path)
    output_file_path = '%s.onnx' % model

    print('Parsing DarkNet cfg file...')
    parser = DarkNetParser()
    layer_configs = parser.parse_cfg_file(cfg_file_path)
    category_num = get_category_num(cfg_file_path)
    output_tensor_names = get_output_convs(layer_configs)
    # e.g. ['036_convolutional', '044_convolutional', '052_convolutional']

    c = (category_num + 5) * get_anchor_num(cfg_file_path)
    h, w = get_h_and_w(layer_configs)
    if len(output_tensor_names) == 2:
        output_tensor_shapes = [
            [c, h // 32, w // 32], [c, h // 16, w // 16]]
    elif len(output_tensor_names) == 3:
        output_tensor_shapes = [
            [c, h // 32, w // 32], [c, h // 16, w // 16],
            [c, h // 8, w // 8]]
    elif len(output_tensor_names) == 4:
        output_tensor_shapes = [
            [c, h // 64, w // 64], [c, h // 32, w // 32],
            [c, h // 16, w // 16], [c, h // 8, w // 8]]
    if is_pan_arch(cfg_file_path):
        output_tensor_shapes.reverse()
    output_tensor_dims = OrderedDict(
        zip(output_tensor_names, output_tensor_shapes))

    print('Building ONNX graph...')
    builder = GraphBuilderONNX(
        model, output_tensor_dims, MAX_BATCH_SIZE)
    yolo_model_def = builder.build_onnx_graph(
        layer_configs=layer_configs,
        weights_file_path=weights_file_path,
        verbose=True)

    print('Checking ONNX model...')
    onnx.checker.check_model(yolo_model_def)

    print('Saving ONNX file...')
    onnx.save(yolo_model_def, output_file_path)

    # print('Done.')



import copy
import torch


def _default_condition(x):
    return isinstance(x, torch.Tensor) and (x.dtype is torch.half or x.dtype is torch.float or x.dtype == torch.bool)


def _make_schema_from_value(value, condition=_default_condition, size=0):
    if condition(value):
        return size, size + 1
    elif isinstance(value, list) or isinstance(value, tuple):
        schema = []
        for child_value in value:
            child_schema, size = _make_schema_from_value(child_value, condition, size)
            schema.append(child_schema)
        if isinstance(value, tuple):
            schema = tuple(schema)
        return schema, size
    elif isinstance(value, dict):
        schema = {}
        for child_key in sorted(value.keys()):
            child_value = value[child_key]
            child_schema, size = _make_schema_from_value(child_value, condition, size)
            schema[child_key] = child_schema
        return schema, size
    else:
        return None, size


class Flattener(object):
    
    def __init__(self, schema, size):
        self._schema = schema
        self._size = size

    @staticmethod
    def from_value(value, condition=_default_condition):
        return Flattener(*_make_schema_from_value(value, condition))
    
    @staticmethod
    def from_dict(x):
        return Flattener(x['schema'], x['size'])

    def dict(self):
        return {'schema': self.schema, 'size': self.size}

    @property
    def schema(self):
        return self._schema

    @property
    def size(self):
        return self._size

    def __len__(self):
        return self._size

    def _flatten(self, value, result):
        if isinstance(self._schema, int):
            result[self._schema] = value
        elif isinstance(self._schema, list) or isinstance(self._schema, tuple):
            for child_value, child_schema in zip(value, self._schema):
                Flattener(child_schema, self.size)._flatten(child_value, result)
        elif isinstance(self._schema, dict):
            for key in sorted(self._schema.keys()):
                child_value = value[key]
                child_schema = self._schema[key]
                Flattener(child_schema, self.size)._flatten(child_value, result)

    def flatten(self, value):
        result = [None for i in range(self.size)]
        self._flatten(value, result)
        return result

    def unflatten(self, flattened):
        if isinstance(self._schema, int):
            return flattened[self._schema]
        elif isinstance(self._schema, list) or isinstance(self._schema, tuple):
            result = []
            for child_schema in self._schema:
                result.append(Flattener(child_schema, self.size).unflatten(flattened))
            if isinstance(self._schema, tuple):
                result = tuple(result)
            return result
        elif isinstance(self._schema, dict):
            result = {}
            for child_key in sorted(self._schema.keys()):
                child_schema = self._schema[child_key]
                result[child_key] = Flattener(child_schema, self.size).unflatten(flattened)
            return result
        else:
            return None